@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -45,6 +45,7 @@
 #include "spc700.h"
 #include "apu.h"
 #include "soundux.h"
+#include "cpuexec.h"
 
 #ifdef DEBUGGER
 FILE *apu_trace = NULL;
@@ -202,9 +203,6 @@ static uint8 ModesToBytes [] = {
 
 extern int NoiseFreq [32];
 
-void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2);
-void StartSample (int channel);
-
 #ifdef DEBUGGER
 static FILE *SoundTracing = NULL;
 
@@ -354,7 +352,7 @@ void S9xPrintAPUState ()
 	    printf ("muted by user using channel on/off toggle\n");
 	else
 	{
-	    int freq = ch->frequency;
+	    int freq = ch->hertz;
 	    if (APU.DSP [APU_NON] & (1 << J)) //ch->type == SOUND_NOISE)
 	    {
 		freq = NoiseFreq [APU.DSP [APU_FLG] & 0x1f];
@@ -375,9 +373,13 @@ void S9xPrintAPUState ()
 	    printf ("left: %d, right: %d, ",
 		    ch->volume_left, ch->volume_right);
 
-	    static char* envelope [] = {"silent", "attack", "decay", "sustain", "release" };
-	    printf ("%s envx: %d, target: %d", ch->state > 4 ? "???" : envelope [ch->state],
-		    ch->envx, ch->envx_target);
+	    static char* envelope [] = 
+	    {
+		"silent", "attack", "decay", "sustain", "release", "gain",
+		"inc_lin", "inc_bent", "dec_lin", "dec_exp"
+	    };
+	    printf ("%s envx: %d, target: %d, %d", ch->state > 9 ? "???" : envelope [ch->state],
+		    ch->envx, ch->envx_target, ch->erate);
 	    printf ("\n");
 	}
     }
@@ -413,9 +415,9 @@ void S9xResetAPU ()
     ZeroMemory (APU.OutPorts, 4);
     IAPU.DirectPage = IAPU.RAM;
     memmove (&IAPU.RAM [0xffc0], APUROM, sizeof (APUROM));
-    memmove (&APU.ExtraRAM, APUROM, sizeof (APUROM));
+    memmove (APU.ExtraRAM, APUROM, sizeof (APUROM));
     IAPU.PC = IAPU.RAM + IAPU.RAM [0xfffe] + (IAPU.RAM [0xffff] << 8);
-    APU.Cycles = Settings.SPCTo65c816Ratio;
+    APU.Cycles = 0;
     APURegisters.YA.W = 0;
     APURegisters.X = 0;
     APURegisters.S = 0xff;
@@ -431,7 +433,9 @@ void S9xResetAPU ()
     APU.ShowROM = TRUE;
     IAPU.RAM [0xf1] = 0x80;
 
-    for (int i = 0; i < 3; i++)
+    int i;
+
+    for (i = 0; i < 3; i++)
     {
 	APU.TimerEnabled [i] = FALSE;
 	APU.TimerValueWritten [i] = 0;
@@ -441,7 +445,12 @@ void S9xResetAPU ()
     for (int j = 0; j < 0x80; j++)
 	APU.DSP [j] = 0;
 
-    APU.DSP [APU_ENDX] = 0xff;
+    IAPU.TwoCycles = IAPU.OneCycle * 2;
+
+    for (i = 0; i < 256; i++)
+	S9xAPUCycles [i] = S9xAPUCycleLengths [i] * IAPU.OneCycle;
+
+    APU.DSP [APU_ENDX] = 0;
     APU.DSP [APU_KOFF] = 0;
     APU.DSP [APU_KON] = 0;
     APU.DSP [APU_FLG] = APU_MUTE | APU_ECHO_DISABLED;
@@ -453,7 +462,8 @@ void S9xResetAPU ()
 
 void S9xSetAPUDSP (uint8 byte)
 {
-    uint8 reg = IAPU.RAM [0xf2] & 0x7f;
+    uint8 reg = IAPU.RAM [0xf2];
+    int i;
 
     switch (reg)
     {
@@ -461,13 +471,13 @@ void S9xSetAPUDSP (uint8 byte)
 	if (byte & APU_SOFT_RESET)
 	{
 	    APU.DSP [reg] = APU_MUTE | APU_ECHO_DISABLED | (byte & 0x1f);
-	    APU.DSP [APU_ENDX] = 0xff;
+	    APU.DSP [APU_ENDX] = 0;
 	    APU.DSP [APU_KOFF] = 0;
 	    APU.DSP [APU_KON] = 0;
 	    S9xSetEchoWriteEnable (FALSE);
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] DSP reset\n", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] DSP reset\n", ICPU.Scanline);
 #endif
 	    // Kill sound
 	    S9xResetSound (FALSE);
@@ -475,18 +485,23 @@ void S9xSetAPUDSP (uint8 byte)
 	else
 	{
 	    S9xSetEchoWriteEnable (!(byte & APU_ECHO_DISABLED));
-
 	    if (byte & APU_MUTE)
 	    {
 #ifdef DEBUGGER
 		if (Settings.TraceSoundDSP)
-		    S9xTraceSoundDSP ("[%d %d] Mute sound\n", CPU.V_Counter,
-				   CPU.Cycles);
+		    S9xTraceSoundDSP ("[%d] Mute sound\n", ICPU.Scanline);
 #endif
 		S9xSetSoundMute (TRUE);
 	    }
 	    else
 		S9xSetSoundMute (FALSE);
+
+	    SoundData.noise_hertz = NoiseFreq [byte & 0x1f];
+	    for (i = 0; i < 8; i++)
+	    {
+		if (SoundData.channels [i].type == SOUND_NOISE)
+		    S9xSetSoundFrequency (i, SoundData.noise_hertz);
+	    }
 	}
 	break;
     case APU_NON:
@@ -494,7 +509,7 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Noise:", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] Noise:", ICPU.Scanline);
 #endif
 	    uint8 mask = 1;
 	    for (int c = 0; c < 8; c++, mask <<= 1)
@@ -537,35 +552,35 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Master volume left:%d\n", 
-			       CPU.V_Counter, CPU.Cycles, (signed char) byte);
+		S9xTraceSoundDSP ("[%d] Master volume left:%d\n", 
+				  ICPU.Scanline, (signed char) byte);
 #endif
-	    S9xSetMasterVolume ((signed char) byte,
-			     (signed char) APU.DSP [APU_MVOL_RIGHT]);
+		S9xSetMasterVolume ((signed char) byte,
+				    (signed char) APU.DSP [APU_MVOL_RIGHT]);
 	}
 	break;
     case APU_MVOL_RIGHT:
 	if (byte != APU.DSP [APU_MVOL_RIGHT])
 	{
-	    S9xSetMasterVolume ((signed char) APU.DSP [APU_MVOL_LEFT],
-			     (signed char) byte);
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Master volume right:%d\n",
-			       CPU.V_Counter, CPU.Cycles, (signed char) byte);
+		S9xTraceSoundDSP ("[%d] Master volume right:%d\n",
+				  ICPU.Scanline, (signed char) byte);
 #endif
+		S9xSetMasterVolume ((signed char) APU.DSP [APU_MVOL_LEFT],
+				    (signed char) byte);
 	}
 	break;
     case APU_EVOL_LEFT:
 	if (byte != APU.DSP [APU_EVOL_LEFT])
 	{
 #ifdef DEBUGGER
-	if (Settings.TraceSoundDSP)
-	    S9xTraceSoundDSP ("[%d %d] Echo volume left:%d\n",
-			   CPU.V_Counter, CPU.Cycles, (signed char) byte);
+	    if (Settings.TraceSoundDSP)
+		S9xTraceSoundDSP ("[%d] Echo volume left:%d\n",
+				  ICPU.Scanline, (signed char) byte);
 #endif
-	    S9xSetEchoVolume ((signed char) byte,
-			      (signed char) APU.DSP [APU_EVOL_RIGHT]);
+		S9xSetEchoVolume ((signed char) byte,
+				  (signed char) APU.DSP [APU_EVOL_RIGHT]);
 	}
 	break;
     case APU_EVOL_RIGHT:
@@ -573,8 +588,8 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Echo volume right:%d\n",
-			       CPU.V_Counter, CPU.Cycles, (signed char) byte);
+		S9xTraceSoundDSP ("[%d] Echo volume right:%d\n",
+				  ICPU.Scanline, (signed char) byte);
 #endif
 		S9xSetEchoVolume ((signed char) APU.DSP [APU_EVOL_LEFT],
 				  (signed char) byte);
@@ -583,8 +598,7 @@ void S9xSetAPUDSP (uint8 byte)
     case APU_ENDX:
 #ifdef DEBUGGER
 	if (Settings.TraceSoundDSP)
-	    S9xTraceSoundDSP ("[%d %d] S9xReset ENDX\n", CPU.V_Counter,
-			   CPU.Cycles);
+	    S9xTraceSoundDSP ("[%d] S9xReset ENDX\n", ICPU.Scanline);
 #endif
 	byte = 0;
 	break;
@@ -595,7 +609,7 @@ void S9xSetAPUDSP (uint8 byte)
 	    uint8 mask = 1;
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Key off:", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] Key off:", ICPU.Scanline);
 #endif
 	    for (int c = 0; c < 8; c++, mask <<= 1)
 	    {
@@ -608,10 +622,12 @@ void S9xSetAPUDSP (uint8 byte)
 #endif		    
 		    if (APU.KeyedChannels & mask)
 		    {
-			APU.KeyedChannels &= ~mask;
-			APU.DSP [APU_KON] &= ~mask;
-			APU.DSP [APU_KOFF] |= mask;
-			S9xSetSoundKeyOff (c);
+			{
+			    APU.KeyedChannels &= ~mask;
+			    APU.DSP [APU_KON] &= ~mask;
+			    //APU.DSP [APU_KOFF] |= mask;
+			    S9xSetSoundKeyOff (c);
+			}
 		    }
 		}
 	    }
@@ -629,7 +645,7 @@ void S9xSetAPUDSP (uint8 byte)
 #ifdef DEBUGGER
 
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] Key on:", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] Key on:", ICPU.Scanline);
 #endif
 	    for (int c = 0; c < 8; c++, mask <<= 1)
 	    {
@@ -639,9 +655,13 @@ void S9xSetAPUDSP (uint8 byte)
 		    if (Settings.TraceSoundDSP)
 			S9xTraceSoundDSP ("%d,", c);
 #endif		    
+		    // Pac-In-Time requires that channels can be key-on
+		    // regardeless of their current state.
 		    APU.KeyedChannels |= mask;
 		    APU.DSP [APU_KON] |= mask;
-		    StartSample (c);
+		    APU.DSP [APU_KOFF] &= ~mask;
+		    APU.DSP [APU_ENDX] &= ~mask;
+		    S9xPlaySample (c);
 		}
 	    }
 #ifdef DEBUGGER
@@ -663,11 +683,11 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d volume left: %d\n", 
-			       CPU.V_Counter, CPU.Cycles, reg>>4, (signed char) byte);
+		S9xTraceSoundDSP ("[%d] %d volume left: %d\n", 
+				  ICPU.Scanline, reg>>4, (signed char) byte);
 #endif
-	    S9xSetSoundVolume (reg >> 4, (signed char) byte,
-			       (signed char) APU.DSP [reg + 1]);
+		S9xSetSoundVolume (reg >> 4, (signed char) byte,
+				   (signed char) APU.DSP [reg + 1]);
 	}
 	break;
     case APU_VOL_RIGHT + 0x00:
@@ -682,11 +702,11 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d volume right: %d\n", 
-			       CPU.V_Counter, CPU.Cycles, reg>>4, (signed char) byte);
+		S9xTraceSoundDSP ("[%d] %d volume right: %d\n", 
+				  ICPU.Scanline, reg >>4, (signed char) byte);
 #endif
-	    S9xSetSoundVolume (reg >> 4, (signed char) APU.DSP [reg - 1],
-			       (signed char) byte);
+		S9xSetSoundVolume (reg >> 4, (signed char) APU.DSP [reg - 1],
+				   (signed char) byte);
 	}
 	break;
 
@@ -700,10 +720,10 @@ void S9xSetAPUDSP (uint8 byte)
     case APU_P_LOW + 0x70:
 #ifdef DEBUGGER
 	if (Settings.TraceSoundDSP)
-	    S9xTraceSoundDSP ("[%d %d] %d freq low: %d\n",
-			   CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+	    S9xTraceSoundDSP ("[%d] %d freq low: %d\n",
+			      ICPU.Scanline, reg>>4, byte);
 #endif
-	S9xSetSoundFrequency (reg >> 4, ((byte + (APU.DSP [reg + 1] << 8)) & FREQUENCY_MASK) * 8);
+	    S9xSetSoundHertz (reg >> 4, ((byte + (APU.DSP [reg + 1] << 8)) & FREQUENCY_MASK) * 8);
 	break;
 
     case APU_P_HIGH + 0x00:
@@ -716,10 +736,11 @@ void S9xSetAPUDSP (uint8 byte)
     case APU_P_HIGH + 0x70:
 #ifdef DEBUGGER
 	if (Settings.TraceSoundDSP)
-	    S9xTraceSoundDSP ("[%d %d] %d freq high: %d\n",
-			   CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+	    S9xTraceSoundDSP ("[%d] %d freq high: %d\n",
+			      ICPU.Scanline, reg>>4, byte);
 #endif
-	S9xSetSoundFrequency (reg >> 4, (((byte << 8) + APU.DSP [reg - 1]) & FREQUENCY_MASK) * 8);
+	    S9xSetSoundHertz (reg >> 4, 
+			      (((byte << 8) + APU.DSP [reg - 1]) & FREQUENCY_MASK) * 8);
 	break;
 
     case APU_SRCN + 0x00:
@@ -734,8 +755,8 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d sample number: %d\n",
-			       CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+		S9xTraceSoundDSP ("[%d] %d sample number: %d\n",
+				  ICPU.Scanline, reg>>4, byte);
 #endif
 	    S9xSetSoundSample (reg >> 4, byte);
 	}
@@ -753,10 +774,13 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d adsr1: %02x\n",
-			       CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+		S9xTraceSoundDSP ("[%d] %d adsr1: %02x\n",
+				  ICPU.Scanline, reg>>4, byte);
 #endif
-	    FixEnvelope (reg >> 4, APU.DSP [reg + 2], byte, APU.DSP [reg + 1]);
+	    {
+		S9xFixEnvelope (reg >> 4, APU.DSP [reg + 2], byte, 
+			     APU.DSP [reg + 1]);
+	    }
 	}
 	break;
 
@@ -772,10 +796,13 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d adsr2: %02x\n", 
-			       CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+		S9xTraceSoundDSP ("[%d] %d adsr2: %02x\n", 
+				  ICPU.Scanline, reg>>4, byte);
 #endif
-	    FixEnvelope (reg >> 4, APU.DSP [reg + 1], APU.DSP [reg - 1], byte);
+	    {
+		S9xFixEnvelope (reg >> 4, APU.DSP [reg + 1], APU.DSP [reg - 1],
+			     byte);
+	    }
 	}
 	break;
 
@@ -791,10 +818,13 @@ void S9xSetAPUDSP (uint8 byte)
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
-		S9xTraceSoundDSP ("[%d %d] %d gain: %02x\n",
-			       CPU.V_Counter, CPU.Cycles, reg>>4, byte);
+		S9xTraceSoundDSP ("[%d] %d gain: %02x\n",
+				  ICPU.Scanline, reg>>4, byte);
 #endif
-	    FixEnvelope (reg >> 4, byte, APU.DSP [reg - 2], APU.DSP [reg - 1]);
+	    {
+		S9xFixEnvelope (reg >> 4, byte, APU.DSP [reg - 2],
+			     APU.DSP [reg - 1]);
+	    }
 	}
 	break;
 
@@ -821,18 +851,18 @@ void S9xSetAPUDSP (uint8 byte)
     case APU_DIR:
 #ifdef DEBUGGER
 	if (Settings.TraceSoundDSP)
-	    S9xTraceSoundDSP ("[%d %d] Sample directory to: %02x\n",
-			   CPU.V_Counter, CPU.Cycles, byte);
+	    S9xTraceSoundDSP ("[%d] Sample directory to: %02x\n",
+			      ICPU.Scanline, byte);
 #endif
 	break;
 
     case APU_PMON:
-	if (byte != APU.DSP [reg])
+	if (byte != APU.DSP [APU_PMON])
 	{
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
 	    {
-		S9xTraceSoundDSP ("[%d %d] FreqMod:", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] FreqMod:", ICPU.Scanline);
 		uint8 mask = 1;
 		for (int c = 0; c < 8; c++, mask <<= 1)
 		{
@@ -852,7 +882,7 @@ void S9xSetAPUDSP (uint8 byte)
 		S9xTraceSoundDSP ("\n");
 	    }
 #endif
-	    S9xSetFrequencyModulationEnable (byte);
+		S9xSetFrequencyModulationEnable (byte);
 	}
 	break;
 
@@ -862,7 +892,7 @@ void S9xSetAPUDSP (uint8 byte)
 #ifdef DEBUGGER
 	    if (Settings.TraceSoundDSP)
 	    {
-		S9xTraceSoundDSP ("[%d %d] Echo:", CPU.V_Counter, CPU.Cycles);
+		S9xTraceSoundDSP ("[%d] Echo:", ICPU.Scanline);
 		uint8 mask = 1;
 		for (int c = 0; c < 8; c++, mask <<= 1)
 		{
@@ -882,7 +912,7 @@ void S9xSetAPUDSP (uint8 byte)
 		S9xTraceSoundDSP ("\n");
 	    }
 #endif
-	    S9xSetEchoEnable (byte);
+		S9xSetEchoEnable (byte);
 	}
 	break;
 
@@ -908,20 +938,23 @@ void S9xSetAPUDSP (uint8 byte)
 	S9xSetFilterCoefficient (reg >> 4, (signed char) byte);
 	break;
     default:
+// XXX
+//printf ("Write %02x to unknown APU register %02x\n", byte, reg);
 	break;
     }
 
-    APU.DSP [reg] = byte;
+    if (reg < 0x80)
+	APU.DSP [reg] = byte;
 }
 
-void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2)
+void S9xFixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2)
 {
     if (adsr1 & 0x80)
     {
 	// ADSR mode
 	static unsigned long AttackRate [16] = {
 	    4100, 2600, 1500, 1000, 640, 380, 260, 160,
-	    96, 64, 40, 24, 16, 10, 6, 0
+	    96, 64, 40, 24, 16, 10, 6, 1
 	};
 	static unsigned long DecayRate [8] = {
 	    1200, 740, 440, 290, 180, 110, 74, 37
@@ -932,12 +965,23 @@ void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2)
 	    1200, 880, 740, 590, 440, 370, 290, 220,
 	    180, 150, 110, 92, 74, 55, 37, 18
 	};
-	if (S9xSetSoundMode (channel, SOUND_MODE_ADSR) == SOUND_MODE_ADSR)
+	// XXX: can DSP be switched to ADSR mode directly from GAIN/INCREASE/
+	// DECREASE mode? And if so, what stage of the sequence does it start
+	// at?
+	if (S9xSetSoundMode (channel, MODE_ADSR))
 	{
-	    S9xSetSoundADSR (channel, AttackRate [adsr1 & 0xf],
-			  DecayRate [(adsr1 >> 4) & 7],
-			  SustainRate [adsr2 & 0x1f],
-			  (adsr2 >> 5) & 7, 8);
+	    // Hack for ROMs that use a very short attack rate, key on a 
+	    // channel, then switch to decay mode. e.g. Final Fantasy II.
+
+	    int attack = AttackRate [adsr1 & 0xf];
+
+	    if (attack == 1 && !Settings.SoundSync)
+		attack = 0;
+
+	    S9xSetSoundADSR (channel, attack,
+			     DecayRate [(adsr1 >> 4) & 7],
+			     SustainRate [adsr2 & 0x1f],
+			     (adsr2 >> 5) & 7, 8);
 	}
     }
     else
@@ -945,9 +989,9 @@ void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2)
 	// Gain mode
 	if ((gain & 0x80) == 0)
 	{
-	    if (S9xSetSoundMode (channel, SOUND_MODE_GAIN) == SOUND_MODE_GAIN)
+	    if (S9xSetSoundMode (channel, MODE_GAIN))
 	    {
-		S9xSetSoundADSR (channel, 0, 0, ~0, 7, 8);
+		S9xSetEnvelopeRate (channel, 0, 0, gain & 0x7f);
 		S9xSetEnvelopeHeight (channel, gain & 0x7f);
 	    }
 	}
@@ -968,24 +1012,23 @@ void FixEnvelope (int channel, uint8 gain, uint8 adsr1, uint8 adsr2)
 	    if (gain & 0x40)
 	    {
 		// Increase mode
-		// XXX: Bent curve or straight line
-		if (S9xSetSoundMode (channel, SOUND_MODE_INCREASE) == SOUND_MODE_INCREASE)
+		if (S9xSetSoundMode (channel, (gain & 0x20) ?
+					  MODE_INCREASE_BENT_LINE :
+					  MODE_INCREASE_LINEAR))
 		{
-		    S9xSetSoundADSR (channel, IncreaseRate [gain & 0x1f],
-				  ~0, ~0, 7, 8);
+		    S9xSetEnvelopeRate (channel, IncreaseRate [gain & 0x1f],
+					1, 127);
 		}
 	    }
 	    else
 	    {
-		// XXX: Bent curve or straight line
-		// Switch to decrease mode from ADSR mode
-		if (S9xSetSoundMode (channel, SOUND_MODE_DECREASE) == SOUND_MODE_DECREASE)
-		{
-		    S9xSetSoundADSR (channel, 0,
-				  (gain & 0x20) != 0 ? DecreaseRateExp [gain & 0x1f] :
-				  IncreaseRate [gain & 0x1f], ~0, 7, 8);
-		    S9xSetSoundDecayMode (channel);
-		}
+		uint32 rate = (gain & 0x20) ? DecreaseRateExp [gain & 0x1f] / 2 :
+					      IncreaseRate [gain & 0x1f];
+		int mode = (gain & 0x20) ? MODE_DECREASE_EXPONENTIAL
+					 : MODE_DECREASE_LINEAR;
+
+		if (S9xSetSoundMode (channel, mode))
+		    S9xSetEnvelopeRate (channel, rate, -1, 0);
 	    }
 	}
     }
@@ -1025,9 +1068,21 @@ void S9xSetAPUControl (uint8 byte)
 	IAPU.RAM [0xF6] = IAPU.RAM [0xF7] = 0;
 
     if (byte & 0x80)
-	APU.ShowROM = TRUE;
+    {
+	if (!APU.ShowROM)
+	{
+	    memmove (&IAPU.RAM [0xffc0], APUROM, sizeof (APUROM));
+	    APU.ShowROM = TRUE;
+	}
+    }
     else
-	APU.ShowROM = FALSE;
+    {
+	if (APU.ShowROM)
+	{
+	    APU.ShowROM = FALSE;
+	    memmove (&IAPU.RAM [0xffc0], APU.ExtraRAM, sizeof (APUROM));
+	}
+    }
     IAPU.RAM [0xf1] = byte;
 }
 
@@ -1099,50 +1154,3 @@ uint8 S9xGetAPUDSP ()
     }
     return (byte);
 }
-
-void S9xAPUSetEndOfSample (int ch)
-{
-    APU.DSP [APU_ENDX] |= 1 << ch;
-    APU.DSP [APU_KON] &= ~(1 << ch);
-    APU.DSP [APU_KOFF] &= ~(1 << ch);
-}
-#ifdef __DJGPP
-END_OF_FUNCTION (S9xAPUSetEndOfSample)
-#endif
-
-void S9xAPUSetEndX (int ch)
-{
-    APU.DSP [APU_ENDX] |= 1 << ch;
-}
-#ifdef __DJGPP
-END_OF_FUNCTION (S9xAPUSetEndX)
-#endif
-
-void StartSample (int c)
-{
-    // Convert sample and start playing sound
-    uint16 sample_number = APU.DSP [APU_SRCN + c * 0x10];
-    int freq;
-    int type;
-    if (APU.DSP [APU_NON] & (1 << c))
-    {
-	type = SOUND_NOISE;
-	freq = NoiseFreq [APU.DSP [APU_FLG] & 0x1f];
-    }
-    else
-    {
-	type = SOUND_SAMPLE;
-	freq = ((APU.DSP [APU_P_LOW + c * 0x10] +
-		(APU.DSP [APU_P_HIGH + c * 0x10] << 8)) & FREQUENCY_MASK) * 8;
-    }
-    S9xSetSoundMode (c, SOUND_MODE_NONE);
-    FixEnvelope (c, 
-		 APU.DSP [APU_GAIN  + (c << 4)], 
-		 APU.DSP [APU_ADSR1 + (c << 4)],
-		 APU.DSP [APU_ADSR2 + (c << 4)]);
-
-    S9xPlaySample (c, type,
-		   (signed char) APU.DSP [APU_VOL_LEFT + c * 0x10],
-		   (signed char) APU.DSP [APU_VOL_RIGHT + c * 0x10],
-		   freq, sample_number);
-}
