@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -52,6 +52,7 @@
 #include "missing.h"
 #include "display.h"
 #include "apu.h"
+#include "sa1.h"
 
 /**********************************************************************************************/
 /* Debug Globals                                                                              */
@@ -61,7 +62,7 @@
 static void WhatsMissing ();
 static void WhatsUsed ();
 EXTERN_C SDMA DMA[8];
-extern int OBJList[4][129];
+extern int OBJList[129];
 
 FILE *trace = NULL;
 struct SBreakPoint S9xBreakpoint[6];
@@ -179,10 +180,7 @@ int AddrModes[256] = {
     2, 10, 3, 19, 6, 6, 6, 12, 0,  1, 0, 0, 14, 14, 14, 17, //E
     4, 11, 9, 20,14, 7, 7, 13, 0, 16, 0, 0, 23, 15, 15, 18  //F
 };
-/**********************************************************************************************/
-/* S9xOPrint()                                                                                   */
-/* This function prints a opcode line in a string for the selected address                    */
-/**********************************************************************************************/
+
 uint8 S9xOPrint (char *Line, uint8 Bank, uint16 Address)
 {
     uint8 S9xOpcode;
@@ -600,6 +598,422 @@ uint8 S9xOPrint (char *Line, uint8 Bank, uint16 Address)
     return Size;
 #endif
 }
+
+uint8 S9xSA1OPrint (char *Line, uint8 Bank, uint16 Address)
+{
+    uint8 S9xOpcode;
+    uint8 Operant[3];
+    uint16 Word;
+    uint8 Byte;
+    uint8 Size = 0;
+    char SByte;
+    short SWord;
+
+#if 0
+    sprintf (Line, "%04X%04X%04X%04X%02X%04X%c%c%c%c%c%c%c%c%c%03d%03d",
+	     SA1Registers.A.W, SA1Registers.X.W, SA1Registers.Y.W,
+	     SA1Registers.D.W, SA1Registers.DB, SA1Registers.S.W,
+	     SA1CheckEmulation () ? 'E' : 'e',
+	     SA1CheckNegative () ? 'N' : 'n',
+	     SA1CheckOverflow () ? 'V' : 'v',
+	     SA1CheckMemory () ? 'M' : 'm',
+	     SA1CheckIndex () ? 'X' : 'x',
+	     SA1CheckDecimal () ? 'D' : 'd',
+	     SA1CheckIRQ () ? 'I' : 'i',
+	     SA1CheckZero () ? 'Z' : 'z',
+	     SA1CheckCarry () ? 'C' : 'c',
+	     CPU.Cycles,
+	     CPU.V_Counter);
+    return (0);
+
+#else
+    S9xOpcode = S9xSA1GetByte ((Bank << 16) + Address);
+    sprintf (Line, "$%02X:%04X %02X ", Bank, Address, S9xOpcode);
+    Operant[0] = S9xSA1GetByte ((Bank << 16) + Address + 1);
+    Operant[1] = S9xSA1GetByte ((Bank << 16) + Address + 2);
+    Operant[2] = S9xSA1GetByte ((Bank << 16) + Address + 3);
+
+    switch (AddrModes[S9xOpcode])
+    {
+    case 0:
+	//Implied
+	    sprintf (Line, "%s         %s", Line, S9xMnemonics[S9xOpcode]);
+	Size = 1;
+	break;
+    case 1:
+	//Immediate[MemoryFlag]
+	    if (!SA1CheckFlag (MemoryFlag))
+	{
+	    //Accumulator 16 - Bit
+		sprintf (Line, "%s%02X %02X    %s #$%02X%02X",
+			 Line,
+			 Operant[0],
+			 Operant[1],
+			 S9xMnemonics[S9xOpcode],
+			 Operant[1],
+			 Operant[0]);
+	    Size = 3;
+	}
+	else
+	{
+	    //Accumulator 8 - Bit
+		sprintf (Line, "%s%02X       %s #$%02X",
+			 Line,
+			 Operant[0],
+			 S9xMnemonics[S9xOpcode],
+			 Operant[0]);
+	    Size = 2;
+	}
+	break;
+    case 2:
+	//Immediate[IndexFlag]
+	    if (!SA1CheckFlag (IndexFlag))
+	{
+	    //X / Y 16 - Bit
+		sprintf (Line, "%s%02X %02X    %s #$%02X%02X",
+			 Line,
+			 Operant[0],
+			 Operant[1],
+			 S9xMnemonics[S9xOpcode],
+			 Operant[1],
+			 Operant[0]);
+	    Size = 3;
+	}
+	else
+	{
+	    //X / Y 8 - Bit
+		sprintf (Line, "%s%02X       %s #$%02X",
+			 Line,
+			 Operant[0],
+			 S9xMnemonics[S9xOpcode],
+			 Operant[0]);
+	    Size = 2;
+	}
+	break;
+    case 3:
+	//Immediate[Always 8 - Bit]
+	    if (1)
+	{
+	    //Always 8 - Bit
+		sprintf (Line, "%s%02X       %s #$%02X",
+			 Line,
+			 Operant[0],
+			 S9xMnemonics[S9xOpcode],
+			 Operant[0]);
+	    Size = 2;
+	}
+	break;
+    case 4:
+	//Relative
+	    sprintf (Line, "%s%02X       %s $%02X",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	SByte = Operant[0];
+	Word = Address;
+	Word += SByte;
+	Word += 2;
+	sprintf (Line, "%-32s[$%04X]", Line, Word);
+	Size = 2;
+	break;
+    case 5:
+	//Relative Long
+	    sprintf (Line, "%s%02X %02X    %s $%02X%02X",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	SWord = (Operant[1] << 8) | Operant[0];
+	Word = Address;
+	Word += SWord;
+	Word += 3;
+	sprintf (Line, "%-32s[$%04X]", Line, Word);
+	Size = 3;
+	break;
+    case 6:
+	//Direct
+	    sprintf (Line, "%s%02X       %s $%02X",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	sprintf (Line, "%-32s[$00:%04X]", Line, Word);
+	Size = 2;
+	break;
+    case 7:
+	//Direct indexed (with x)
+	    sprintf (Line, "%s%02X       %s $%02X,x",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Word += SA1Registers.X.W;
+	sprintf (Line, "%-32s[$00:%04X]", Line, Word);
+	Size = 2;
+	break;
+    case 8:
+	//Direct indexed (with y)
+	    sprintf (Line, "%s%02X       %s $%02X,y",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Word += SA1Registers.Y.W;
+	sprintf (Line, "%-32s[$00:%04X]", Line, Word);
+	Size = 2;
+	break;
+    case 9:
+	//Direct Indirect
+	    sprintf (Line, "%s%02X       %s ($%02X)",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Word = S9xSA1GetWord (Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 2;
+	break;
+    case 10:
+	//Direct Indexed Indirect
+	    sprintf (Line, "%s%02X       %s ($%02X,x)",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Word += SA1Registers.X.W;
+	Word = S9xSA1GetWord (Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 2;
+	break;
+    case 11:
+	//Direct Indirect Indexed
+	    sprintf (Line, "%s%02X       %s ($%02X),y",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Word = S9xSA1GetWord (Word);
+	Word += SA1Registers.Y.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 2;
+	break;
+    case 12:
+	//Direct Indirect Long
+	    sprintf (Line, "%s%02X       %s [$%02X]",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Byte = S9xSA1GetByte (Word + 2);
+	Word = S9xSA1GetWord (Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, Byte, Word);
+	Size = 2;
+	break;
+    case 13:
+	//Direct Indirect Indexed Long
+	    sprintf (Line, "%s%02X       %s [$%02X],y",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = Operant[0];
+	Word += SA1Registers.D.W;
+	Byte = S9xSA1GetByte (Word + 2);
+	Word = S9xSA1GetWord (Word);
+	Word += SA1Registers.Y.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, Byte, Word);
+	Size = 2;
+	break;
+    case 14:
+	//Absolute
+	    sprintf (Line, "%s%02X %02X    %s $%02X%02X",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 3;
+	break;
+    case 15:
+	//Absolute Indexed (With X)
+	    sprintf (Line, "%s%02X %02X    %s $%02X%02X,x",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Word += SA1Registers.X.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 3;
+	break;
+    case 16:
+	//Absolute Indexed (With Y)
+	    sprintf (Line, "%s%02X %02X    %s $%02X%02X,y",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Word += SA1Registers.Y.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 3;
+	break;
+    case 17:
+	//Absolute long
+	 sprintf (Line, "%s%02X %02X %02X %s $%02X%02X%02X",
+		   Line,
+		   Operant[0],
+		   Operant[1],
+		   Operant[2],
+		   S9xMnemonics[S9xOpcode],
+		   Operant[2],
+		   Operant[1],
+		   Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, Operant[2], Word);
+	Size = 4;
+	break;
+    case 18:
+	//Absolute Indexed long
+	 sprintf (Line, "%s%02X %02X %02X %s $%02X%02X%02X,x",
+		   Line,
+		   Operant[0],
+		   Operant[1],
+		   Operant[2],
+		   S9xMnemonics[S9xOpcode],
+		   Operant[2],
+		   Operant[1],
+		   Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Word += SA1Registers.X.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, Operant[2], Word);
+	Size = 4;
+	break;
+    case 19:
+	//StackRelative
+	    sprintf (Line, "%s%02X       %s $%02X,s",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = SA1Registers.S.W;
+	Word += Operant[0];
+	sprintf (Line, "%-32s[$00:%04X]", Line, Word);
+	Size = 2;
+	break;
+    case 20:
+	//Stack Relative Indirect Indexed
+	    sprintf (Line, "%s%02X       %s ($%02X,s),y",
+		     Line,
+		     Operant[0],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[0]);
+	Word = SA1Registers.S.W;
+	Word += Operant[0];
+	Word = S9xSA1GetWord (Word);
+	Word += SA1Registers.Y.W;
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.DB, Word);
+	Size = 2;
+	break;
+    case 21:
+	//Absolute Indirect
+	    sprintf (Line, "%s%02X %02X    %s ($%02X%02X)",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Word = S9xSA1GetWord (Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.PB, Word);
+	Size = 3;
+	break;
+    case 22:
+	//Absolute Indirect Long
+	    sprintf (Line, "%s%02X %02X    %s [$%02X%02X]",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Byte = S9xSA1GetByte (Word + 2);
+	Word = S9xSA1GetWord (Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, Byte, Word);
+	Size = 3;
+	break;
+    case 23:
+	//Absolute Indexed Indirect
+	    sprintf (Line, "%s%02X %02X    %s ($%02X%02X,x)",
+		     Line,
+		     Operant[0],
+		     Operant[1],
+		     S9xMnemonics[S9xOpcode],
+		     Operant[1],
+		     Operant[0]);
+	Word = (Operant[1] << 8) | Operant[0];
+	Word += SA1Registers.X.W;
+	Word = S9xSA1GetWord (SA1.ShiftedPB + Word);
+	sprintf (Line, "%-32s[$%02X:%04X]", Line, SA1Registers.PB, Word);
+	Size = 3;
+	break;
+    case 24:
+	//Implied accumulator
+	    sprintf (Line, "%s         %s A", Line, S9xMnemonics[S9xOpcode]);
+	Size = 1;
+	break;
+    case 25:
+	// MVN/MVP SRC DST
+	    sprintf (Line, "%s         %s %02X %02X", Line, S9xMnemonics[S9xOpcode],
+		     Operant[0], Operant[1]);
+	Size = 3;
+	break;
+    }
+    sprintf (Line, "%-44s A:%04X X:%04X Y:%04X D:%04X DB:%02X S:%04X P:%c%c%c%c%c%c%c%c%c HC:%03d VC:%03ld",
+	     Line, SA1Registers.A.W, SA1Registers.X.W, SA1Registers.Y.W,
+	     SA1Registers.D.W, SA1Registers.DB, SA1Registers.S.W,
+	     SA1CheckEmulation () ? 'E' : 'e',
+	     SA1CheckNegative () ? 'N' : 'n',
+	     SA1CheckOverflow () ? 'V' : 'v',
+	     SA1CheckMemory () ? 'M' : 'm',
+	     SA1CheckIndex () ? 'X' : 'x',
+	     SA1CheckDecimal () ? 'D' : 'd',
+	     SA1CheckIRQ () ? 'I' : 'i',
+	     SA1CheckZero () ? 'Z' : 'z',
+	     SA1CheckCarry () ? 'C' : 'c',
+	     CPU.Cycles,
+	     CPU.V_Counter);
+
+    return Size;
+#endif
+}
+
 /**********************************************************************************************/
 /* DPrint()                                                                                   */
 /* This function prints a line in the debug listbox and deletes upperlines if needed          */
@@ -683,6 +1097,18 @@ void ProcessDebugCommand (char *Line)
 	Address = CPU.PC - CPU.PCBase;
 	Line[0] = 'r';
     }
+    if (Line[0] == 'z')
+    {
+	uint16 *p = (uint16 *) &Memory.VRAM [PPU.BG[2].SCBase << 1];
+	for (int l = 0; l < 32; l++)
+	{
+	    for (int c = 0; c < 32; c++, p++)
+	    {
+		printf ("%04x,", *p++);
+	    }
+	    printf ("\n");
+	}
+    }
     if (*Line == 'c')
     {
 	printf ("Colours:\n");
@@ -730,10 +1156,13 @@ void ProcessDebugCommand (char *Line)
 		PPU.OBJNameSelect, PPU.FirstSprite);
 	for (int p = 0; p < 4; p++)
 	{
-	    int c;
-	    for (c = 0; c < 128; c++)
-		if (OBJList[p][c] < 0)
-		    break;
+	    int c = 0;
+	    int i;
+	    for (i = 0; OBJList [i] >= 0; i++)
+	    {
+		if (PPU.OBJ[OBJList [i]].Priority == p)
+		    c++;
+	    }
 	    printf ("Priority %d: %03d, ", p, c);
 	}
 	printf ("\n");
@@ -752,18 +1181,37 @@ void ProcessDebugCommand (char *Line)
     }
     if (*Line == 'T')
     {
-	CPU.Flags ^= TRACE_FLAG;
-	if (CPU.Flags & TRACE_FLAG)
+	if (Line [1] == 'S')
 	{
-	    printf ("CPU instruction tracing enabled.\n");
-	    if (trace == NULL)
-		trace = fopen ("trace.log", "wb");
+	    SA1.Flags ^= TRACE_FLAG;
+	    if (SA1.Flags & TRACE_FLAG)
+	    {
+		printf ("SA1 CPU instruction tracing enabled.\n");
+		if (trace == NULL)
+		    trace = fopen ("trace.log", "wb");
+	    }
+	    else
+	    {
+		printf ("SA1 CPU instruction tracing disabled.\n");
+		fclose (trace);
+		trace = NULL;
+	    }
 	}
 	else
 	{
-	    printf ("CPU instruction tracing disabled.\n");
-	    fclose (trace);
-	    trace = NULL;
+	    CPU.Flags ^= TRACE_FLAG;
+	    if (CPU.Flags & TRACE_FLAG)
+	    {
+		printf ("CPU instruction tracing enabled.\n");
+		if (trace == NULL)
+		    trace = fopen ("trace.log", "wb");
+	    }
+	    else
+	    {
+		printf ("CPU instruction tracing disabled.\n");
+		fclose (trace);
+		trace = NULL;
+	    }
 	}
     }
     if (*Line == 'A')
@@ -1152,6 +1600,9 @@ static void WhatsUsed ()
 {
     printf ("V-line: %ld, H-Pos: %ld\n", CPU.V_Counter, CPU.Cycles);
     printf ("Screen mode: %d, ", PPU.BGMode);
+    if (PPU.BGMode <= 1 && (Memory.FillRAM [0x2105] & 8))
+	printf ("(BG#2 Priority)");
+    
     printf ("Brightness: %d", PPU.Brightness);
     if (Memory.FillRAM[0x2100] & 0x80)
 	printf (" (screen blanked)");
@@ -1574,9 +2025,34 @@ void S9xTrace ()
     fprintf (trace, "%s\n", String);
 }
 
+void S9xSA1Trace ()
+{
+    char String [512];
+    S9xSA1OPrint (String, SA1Registers.PB, SA1.PC - SA1.PCBase);
+    fprintf (trace, "%s\n", String);
+    fflush (trace);
+}
+
 void S9xTraceMessage (const char *s)
 {
     fprintf (trace, "%s\n", s);
 }
 
+extern "C" void TraceSA1 ()
+{
+    CPU.Flags ^= TRACE_FLAG;
+    if (CPU.Flags & TRACE_FLAG)
+    {
+	printf ("SA1 CPU instruction tracing enabled.\n");
+	if (trace == NULL)
+	    trace = fopen ("trace.log", "wb");
+    }
+    else
+    {
+	printf ("SA1 CPU instruction tracing disabled.\n");
+	fclose (trace);
+	trace = NULL;
+    }
+}
+
 #endif
