@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -42,6 +42,10 @@
 #include "memmap.h"
 #include <math.h>
 
+#ifndef M_PI
+#define M_PI 3.14159265359
+#endif
+
 #define CLIPU8(v) if(v < 0) v = 0; else if(v > 255) v = 255
 #define CLIP8(v) if(v < -128) v = -128; else if(v > 127) v = 127
 #define CLIP16(v) if(v < -32768) v = -32768; else if(v > 32767) v = 32767
@@ -331,7 +335,7 @@ void S9xSetDSP (uint8 byte, uint16 address)
 //		    case 0x80: DSP1.in_count = 2;	break;
 
 		    default:
-printf ("unknown command %02x\n", byte);
+printf ("unknown DSP1 command %02x\n", byte);
 		    case 0x80:
 			DSP1.in_count = 0;
 			DSP1.waiting4command = TRUE;
@@ -692,10 +696,6 @@ printf ("unknown command %02x\n", byte);
 
 ///////////////////////// DSP1 Calculation code /////////////////////////
 
-#ifndef M_PI
-#define M_PI 3.14159265359
-#endif
-
 // Convert DSP1 type angle to a radian
 #define ANGLE_TO_RAD(angle) (double(angle) * M_PI / 128.0)
 
@@ -1468,13 +1468,13 @@ void DSPOp0A ()
 
     if (Op0AVS != 0)
     {
-	Op0AC = int16(((Op02CXF - ((GroundRX + GroundLX) / 2))) / Op0AVS);
-	Op0AD = int16((((Op02CYF - ((GroundRY + GroundLY) / 2))) / Op0AVS) * 256);
+	Op0AC = int16((((Op02CXF - ((GroundRX + GroundLX) / 2.0))) / Op0AVS) * 256.0);
+	Op0AD = int16((((Op02CYF - ((GroundRY + GroundLY) / 2.0))) / Op0AVS) * 256.0);
     }
     else
     {
-	Op0AC = 0;
-	Op0AD = 0;
+	Op0AC = 256;
+	Op0AD = 256;
     }
     Op0AVS += 1;
 }
