@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -42,7 +42,7 @@
 #include <stdio.h>
 #include <string.h>
 
-extern char const * fx_apvMnemonicTable[];
+extern const char *fx_apvMnemonicTable[];
 extern struct FxRegs_s GSU;
 
 
@@ -74,13 +74,13 @@ void FxPipeString(char * pvString)
 {
     char *p;
     uint32 vOpcode = (GSU.vStatusReg & 0x300) | ((uint32)PIPE);
-    char const *m = fx_apvMnemonicTable[vOpcode];
+    const char *m = fx_apvMnemonicTable[vOpcode];
     uint8 vPipe1,vPipe2,vByte1,vByte2;
-    //uint8 vPipeBank = GSU.vPipeAdr >> 16;
+    uint8 vPipeBank = GSU.vPipeAdr >> 16;
 	
     /* The next two bytes after the pipe's address */
-    //vPipe1 = GSU.apvRomBank[vPipeBank][USEX16(GSU.vPipeAdr+1)];
-    //vPipe2 = GSU.apvRomBank[vPipeBank][USEX16(GSU.vPipeAdr+2)];
+    vPipe1 = GSU.apvRomBank[vPipeBank][USEX16(GSU.vPipeAdr+1)];
+    vPipe2 = GSU.apvRomBank[vPipeBank][USEX16(GSU.vPipeAdr+2)];
     
     /* The actual next two bytes to be read */
     vByte1 = PRGBANK(USEX16(R15));
@@ -88,13 +88,13 @@ void FxPipeString(char * pvString)
 
     /* Print ROM address of the pipe */
     sprintf(pvString, "%02x:%04x %02x       ",
-	    USEX8(GSU.vPrgBankReg), USEX16(R15 - 1), USEX8(PIPE));
+	    USEX8(vPipeBank), USEX16(GSU.vPipeAdr), USEX8(PIPE));
     p = &pvString[strlen(pvString)];
  
     /* Check if it's a branch instruction */
     if( PIPE >= 0x05 && PIPE <= 0x0f )
     {
-	sprintf(&pvString[11], "%02x    ", USEX8(vByte1));
+	sprintf(&pvString[11], "%02x    ", USEX8(vPipe1));
 #ifdef BRANCH_DELAY_RELATIVE
 	sprintf(p, m, USEX16(R15 + SEX8(vByte1) + 1 ) );
 #else
@@ -107,7 +107,7 @@ void FxPipeString(char * pvString)
     /* Check for 'ibt', 'lms' or 'sms' */
     else if( PIPE >= 0xa0 && PIPE <= 0xaf )
     {
-	sprintf(&pvString[11], "%02x    ", USEX8(vByte1));
+	sprintf(&pvString[11], "%02x    ", USEX8(vPipe1));
 	if( (GSU.vStatusReg & 0x300) == 0x100 || (GSU.vStatusReg & 0x300) == 0x200 )
 	    sprintf(p, m, USEX16(vByte1) << 1 );
 	else
@@ -119,7 +119,7 @@ void FxPipeString(char * pvString)
     /* Check for 'iwt', 'lm' or 'sm' */
     else if( PIPE >= 0xf0 )
     {
-	sprintf(&pvString[11], "%02x %02x ", USEX8(vByte1), USEX8(vByte2));
+	sprintf(&pvString[11], "%02x %02x ", USEX8(vPipe1), USEX8(vPipe2));
 	sprintf(p, m, USEX8(vByte1) | (USEX16(vByte2)<<8) );
     }
     /* Normal instruction */
@@ -127,7 +127,7 @@ void FxPipeString(char * pvString)
 	strcpy(p, m);
 }
 
-char const * fx_apvMnemonicTable[] =
+const char *fx_apvMnemonicTable[] =
 {
     /*
      * ALT0 Table
