@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -53,10 +53,16 @@
 #include "ppu.h"
 #include "display.h"
 #include "cheats.h"
+#include "apu.h"
+#include "sa1.h"
 
 #ifndef ZSNES_FX
 #include "fxemu.h"
 extern struct FxInit_s SuperFX;
+#else
+START_EXTERN_C
+extern uint8 *SFXPlotTable;
+END_EXTERN_C
 #endif
 
 extern char *rom_filename;
@@ -71,50 +77,55 @@ bool8 CMemory::AllASCII (uint8 *b, int size)
     return (TRUE);
 }
 
-int CMemory::ScoreHiROM ()
+int CMemory::ScoreHiROM (bool8 skip_header)
 {
     int score = 0;
-    if ((Memory.ROM [0xffdc] + (Memory.ROM [0xffdd] << 8) +
-	 Memory.ROM [0xffde] + (Memory.ROM [0xffdf] << 8)) == 0xffff)
+    int o = skip_header ? 0xff00 + 0x200 : 0xff00;
+
+    if ((Memory.ROM [o + 0xdc] + (Memory.ROM [o + 0xdd] << 8) +
+	 Memory.ROM [o + 0xde] + (Memory.ROM [o + 0xdf] << 8)) == 0xffff)
 	score += 2;
 
-    if (Memory.ROM [0xffda] == 0x33)
+    if (Memory.ROM [o + 0xda] == 0x33)
 	score += 2;
-    if ((Memory.ROM [0xffd5] & 0xf) < 4)
+    if ((Memory.ROM [o + 0xd5] & 0xf) < 4)
 	score += 2;
-    if (!(Memory.ROM [0xfffd] & 0x80))
+    if (!(Memory.ROM [o + 0xfd] & 0x80))
 	score -= 4;
     if (CalculatedSize > 1024 * 1024 * 3)
 	score += 4;
-    if ((1 << (Memory.ROM [0xffd7] - 7)) > 48)
+    if ((1 << (Memory.ROM [o + 0xd7] - 7)) > 48)
 	score -= 1;
-    if (!AllASCII (&Memory.ROM [0xffb0], 6))
+    if (!AllASCII (&Memory.ROM [o + 0xb0], 6))
 	score -= 1;
-    if (!AllASCII (&Memory.ROM [0xffc0], ROM_NAME_LEN - 1))
+    if (!AllASCII (&Memory.ROM [o + 0xc0], ROM_NAME_LEN - 1))
 	score -= 1;
 
     return (score);
 }
 
-int CMemory::ScoreLoROM ()
+int CMemory::ScoreLoROM (bool8 skip_header)
 {
     int score = 0;
+    int o = skip_header ? 0x7f00 + 0x200 : 0x7f00;
 
-    if ((Memory.ROM [0x7fdc] + (Memory.ROM [0x7fdd] << 8) +
-	 Memory.ROM [0x7fde] + (Memory.ROM [0x7fdf] << 8)) == 0xffff)
+    if ((Memory.ROM [o + 0xdc] + (Memory.ROM [o + 0xdd] << 8) +
+	 Memory.ROM [o + 0xde] + (Memory.ROM [o + 0xdf] << 8)) == 0xffff)
 	score += 2;
 
-    if (Memory.ROM [0x7fda] == 0x33)
+    if (Memory.ROM [o + 0xda] == 0x33)
 	score += 2;
-    if ((Memory.ROM [0x7fd5] & 0xf) < 4)
+    if ((Memory.ROM [o + 0xd5] & 0xf) < 4)
 	score += 2;
-    if (!(Memory.ROM [0x7ffd] & 0x80))
+    if (CalculatedSize <= 1024 * 1024 * 16)
+	score += 2;
+    if (!(Memory.ROM [o + 0xfd] & 0x80))
 	score -= 4;
-    if ((1 << (Memory.ROM [0x7fd7] - 7)) > 48)
+    if ((1 << (Memory.ROM [o + 0xd7] - 7)) > 48)
 	score -= 1;
-    if (!AllASCII (&Memory.ROM [0x7fb0], 6))
+    if (!AllASCII (&Memory.ROM [o + 0xb0], 6))
 	score -= 1;
-    if (!AllASCII (&Memory.ROM [0x7fc0], ROM_NAME_LEN - 1))
+    if (!AllASCII (&Memory.ROM [o + 0xc0], ROM_NAME_LEN - 1))
 	score -= 1;
 
     return (score);
@@ -154,6 +165,9 @@ bool8 CMemory::Init ()
     VRAM    = new uint8 [0x10000];
     ROM     = new uint8 [MAX_ROM_SIZE + 0x200 + 0x8000];
     FillRAM = NULL;
+#ifdef ZSNES_FX
+    SFXPlotTable = new uint8 [0x100000];
+#endif
 
     IPPU.TileCache [TILE_2BIT] = new uint8 [MAX_2BIT_TILES * 128];
     IPPU.TileCache [TILE_4BIT] = new uint8 [MAX_4BIT_TILES * 128];
@@ -183,8 +197,8 @@ bool8 CMemory::Init ()
 
 #ifndef ZSNES_FX
     SuperFX.pvRegisters = &Memory.FillRAM [0x3000];
-    SuperFX.nRamBanks = 2;
-    SuperFX.pvRam = ROM + 1024 * 1024 * 4;
+    SuperFX.nRamBanks = 1;
+    SuperFX.pvRam = ::SRAM;
     SuperFX.nRomBanks = (2 * 1024 * 1024) / (32 * 1024);
     SuperFX.pvRom = (uint8 *) ROM;
 #endif
@@ -206,6 +220,9 @@ void CMemory::Deinit ()
     delete VRAM;
     delete ROM;
 
+#ifdef ZSNES_FX
+    delete SFXPlotTable;
+#endif
     delete IPPU.TileCache [TILE_2BIT];
     delete IPPU.TileCache [TILE_4BIT];
     delete IPPU.TileCache [TILE_8BIT];
@@ -231,6 +248,7 @@ bool8 CMemory::LoadROM (const char *filename)
     char name [_MAX_FNAME];
     char ext [_MAX_EXT];
     char fname [_MAX_PATH];
+    int i;
 
     CalculatedSize = 0;
 
@@ -238,6 +256,10 @@ again:
     _splitpath (filename, drive, dir, name, ext);
     _makepath (fname, drive, dir, name, ext);
 
+#ifdef _WIN32
+    memmove (&ext [0], &ext[1], 4);
+#endif
+
     unsigned long TotalFileSize = 0;
 
 #ifdef UNZIP_SUPPORT
@@ -285,6 +307,10 @@ again:
 	    {
 		more = TRUE;
 		ext [0]++;
+#ifdef _WIN32
+                memmove (&ext [1], &ext [0], 4);
+                ext [0] = '.';
+#endif
 		_makepath (fname, drive, dir, name, ext);
 	    }
 	    else
@@ -296,6 +322,10 @@ again:
 	    {
 		more = TRUE;
 		name [len - 1]++;
+#ifdef _WIN32
+                memmove (&ext [1], &ext [0], 4);
+                ext [0] = '.';
+#endif
 		_makepath (fname, drive, dir, name, ext);
 	    }
 	    else
@@ -317,12 +347,20 @@ again:
 			"Found multiple ROM file headers (and ignored them).");
     }
 
-    ZeroMemory (ROM + CalculatedSize, MAX_ROM_SIZE - CalculatedSize);
+    ZeroMemory (ROM + TotalFileSize, MAX_ROM_SIZE - TotalFileSize);
     int orig_hi_score, orig_lo_score;
     int hi_score, lo_score;
 
-    orig_hi_score = hi_score = ScoreHiROM ();
-    orig_lo_score = lo_score = ScoreLoROM ();
+    orig_hi_score = hi_score = ScoreHiROM (FALSE);
+    orig_lo_score = lo_score = ScoreLoROM (FALSE);
+
+    if (HeaderCount == 0 &&
+	((hi_score > lo_score && ScoreHiROM (TRUE) > hi_score) ||
+	 (hi_score <= lo_score && ScoreLoROM (TRUE) > lo_score)))
+    {
+	S9xMessage (S9X_INFO, S9X_HEADER_WARNING, 
+		    "Try specifying the -hd command line option if the game doesn't work\n");
+    }
 
     Interleaved = Settings.ForceInterleaved || Settings.ForceInterleaved2;
     if (Settings.ForceLoROM || (!Settings.ForceHiROM && lo_score >= hi_score))
@@ -330,32 +368,40 @@ again:
 	Memory.LoROM = TRUE;
 	Memory.HiROM = FALSE;
 
-	switch (ROM [0x7fd5] & 0xf)
+	// Ignore map type byte if not 0x2x or 0x3x
+	if ((ROM [0x7fd5] & 0xf0) == 0x20 || (ROM [0x7fd5] & 0xf0) == 0x30)
 	{
-	case 1:
-	    Interleaved = TRUE;
-	    break;
-	case 2:
-	    if (!Settings.ForceLoROM)
+	    switch (ROM [0x7fd5] & 0xf)
 	    {
-		Memory.LoROM = FALSE;
-		Memory.HiROM = TRUE;
+	    case 1:
+		Interleaved = TRUE;
+		break;
+	    case 2:
+		if (!Settings.ForceLoROM &&
+		    strncmp ((char *) &ROM [0x7fc0], "SUPER FORMATION SOCCE", 21) != 0)
+		{
+		    Memory.LoROM = FALSE;
+		    Memory.HiROM = TRUE;
+		}
+		break;
+	    case 5:
+		Interleaved = TRUE;
+		Tales = TRUE;
+		break;
 	    }
-	    break;
-	case 5:
-	    Interleaved = TRUE;
-	    Tales = TRUE;
-	    break;
 	}
     }
     else
     {
-	switch (ROM [0xffd5] & 0xf)
+	if ((ROM [0xffd5] & 0xf0) == 0x20 || (ROM [0xffd5] & 0xf0) == 0x30)
 	{
-	case 0:
-	case 3:
-	    Interleaved = TRUE;
-	    break;
+	    switch (ROM [0xffd5] & 0xf)
+	    {
+	    case 0:
+	    case 3:
+		Interleaved = TRUE;
+		break;
+	    }
 	}
 	Memory.LoROM = FALSE;
 	Memory.HiROM = TRUE;
@@ -374,7 +420,7 @@ again:
 	    
 	nblocks = step;
 	uint8 blocks [256];
-	int i;
+
 	if (Tales)
 	{
 	    nblocks = 0x60;
@@ -433,8 +479,8 @@ again:
 	    }
 	}
 	delete tmp;
-	hi_score = ScoreHiROM ();
-	lo_score = ScoreLoROM ();
+	hi_score = ScoreHiROM (FALSE);
+	lo_score = ScoreLoROM (FALSE);
 
 	if ((Memory.HiROM &&
 	     (lo_score >= hi_score || orig_lo_score > hi_score ||
@@ -465,8 +511,8 @@ again:
     Settings.DSP1Master = Settings.ForceDSP1;
     Settings.SuperFX = FALSE;
 
-    ZeroMemory (BlockIsRAM, 0x800);
-    ZeroMemory (BlockIsROM, 0x800);
+    ZeroMemory (BlockIsRAM, MEMMAP_NUM_BLOCKS);
+    ZeroMemory (BlockIsROM, MEMMAP_NUM_BLOCKS);
 
     ::SRAM = SRAM;
 
@@ -477,6 +523,9 @@ again:
 	ROMSpeed = ROM [0xffd5];
 	ROMType = ROM [0xffd6];
 	ROMSize = ROM [0xffd7];
+	ROMChecksum = ROM [0xffde] + (ROM [0xffdf] << 8);
+	ROMComplementChecksum = ROM [0xffdc] + (ROM [0xffdd] << 8);
+
 	// Try to auto-detect the DSP1 chip
 	if (!Settings.ForceNoDSP1 &&
 	    (ROMType & 0xf) >= 3 && (ROMType & 0xf0) == 0)
@@ -486,11 +535,9 @@ again:
 	{
 	    TalesROMMap ();
 	}
-#if 1
-	else if ((ROMSpeed & 0xf) == 2 && 
+	else if ((ROMSpeed & ~0x10) == 0x22 && 
 		 strncmp (ROMName, "Super Street Fighter", 20) != 0)
 	    AlphaROMMap ();
-#endif
 	else
 	    HiROMMap ();
     }
@@ -501,12 +548,17 @@ again:
 	ROMSpeed = ROM [0x7fd5];
 	ROMType = ROM [0x7fd6];
 	ROMSize = ROM [0x7fd7];
+	ROMChecksum = ROM [0x7fde] + (ROM [0x7fdf] << 8);
+	ROMComplementChecksum = ROM [0x7fdc] + (ROM [0x7fdd] << 8);
+
 	strncpy (ROMName, (char *) &ROM[0x7fc0], ROM_NAME_LEN - 1);
 	Settings.SuperFX = Settings.ForceSuperFX;
 
 	if ((ROMType & 0xf0) == 0x10)
 	    Settings.SuperFX = !Settings.ForceNoSuperFX;
 
+	Settings.SA1 = FALSE;
+
 	// Try to auto-detect the DSP1 chip
 	if (!Settings.ForceNoDSP1 &&
 	    (ROMType & 0xf) >= 3 && (ROMType & 0xf0) == 0)
@@ -514,24 +566,68 @@ again:
 
 	if (Settings.SuperFX)
 	{
-	    ::SRAM = ROM + 1024 * 1024 * 4;
+	    //::SRAM = ROM + 1024 * 1024 * 4;
 	    SuperFXROMMap ();
 	    Settings.MultiPlayer5Master = FALSE;
 	    Settings.MouseMaster = FALSE;
 	    Settings.SuperScopeMaster = FALSE;
 	    Settings.DSP1Master = FALSE;
+	    Settings.SA1 = FALSE;
+	}
+	else
+	if (Settings.ForceSA1 ||
+	    (!Settings.ForceNoSA1 && (ROMSpeed & ~0x10) == 0x23 && 
+	     (ROMType & 0xf) > 3 && (ROMType & 0xf0) == 0x30))
+	{
+	    Settings.SA1 = TRUE;
+	    Settings.MultiPlayer5Master = FALSE;
+	    Settings.MouseMaster = FALSE;
+	    Settings.SuperScopeMaster = FALSE;
+	    Settings.DSP1Master = FALSE;
+	    SA1ROMMap ();
 	}
 	else
 	if ((ROMSpeed & ~0x10) == 0x25)
 	    TalesROMMap ();
 	else
-	if ((ROMSpeed & ~0x10) == 0x23 && 
-	    (ROMType & 0xf) > 3 && (ROMType & 0xf0) == 0x30)
-	    SA1ROMMap ();
+	if (strncmp ((char *) &Memory.ROM [0x7fc0], "SOUND NOVEL-TCOOL",17) == 0) 
+	{
+	    LoROM24MBSMap ();
+	    Settings.DSP1Master = FALSE;
+	}
+	else
+	if (strncmp ((char *) &Memory.ROM [0x7fc0], "DERBY STALLION 96",17) == 0)
+	{
+	    LoROM24MBSMap ();
+	    Settings.DSP1Master = FALSE;
+	}
 	else
 	    LoROMMap ();
     }
 
+    int power2 = 0;
+    int size = TotalFileSize;
+
+    while (size >>= 1)
+	power2++;
+
+    size = 1 << power2;
+    int remainder = TotalFileSize - size;
+
+    int sum1 = 0;
+    int sum2 = 0;
+
+    for (i = 0; i < size; i++)
+	sum1 += ROM [i];
+
+    for (i = 0; i < remainder; i++)
+	sum2 += ROM [size + i];
+
+    if (remainder)
+	sum1 += sum2 * (size / remainder);
+
+    sum1 &= 0xffff;
+
     if (Settings.ForceNTSC)
 	Settings.PAL = FALSE;
     else
@@ -576,17 +672,21 @@ again:
 		    ((1 << (Memory.SRAMSize + 3)) * 128) - 1 : 0;
     }
 
+    IAPU.OneCycle = ONE_APU_CYCLE;
+
     ApplyROMFixes ();
     sprintf (ROMName, "%s", Safe (ROMName));
 
-    sprintf (String, "\"%s\" %s, %s, %s, Type: %s, Mode: %s, TV: %s, S-RAM: %s",
-	    ROMName, MapType (),
-	    Size (),
-	    Speed (),
-	    KartContents (),
-	    MapMode (),
-	    TVStandard (),
-	    StaticRAMSize ());
+    sprintf (String, "\"%s\" [%s] %s, %s, Type: %s, Mode: %s, TV: %s, S-RAM: %s",
+	     ROMName,
+	     (ROMChecksum + ROMComplementChecksum != 0xffff ||
+	      ROMChecksum != sum1) ? "bad checksum" : "checksum ok",
+	     MapType (),
+	     Size (),
+	     KartContents (),
+	     MapMode (),
+	     TVStandard (),
+	     StaticRAMSize ());
 
     S9xMessage (S9X_INFO, S9X_ROM_INFO, String);
 	
@@ -651,7 +751,7 @@ void CMemory::FixROMSpeed ()
 {
     int c;
 
-    for (c = 0x400; c < 0x800; c++)
+    for (c = 0x800; c < 0x1000; c++)
     {
 	if (BlockIsROM [c])
 	    MemorySpeed [c] = CPU.FastROMSpeed;
@@ -661,7 +761,7 @@ void CMemory::FixROMSpeed ()
 void CMemory::WriteProtectROM ()
 {
     memmove ((void *) WriteMap, (void *) Map, sizeof (Map));
-    for (int c = 0; c < 0x800; c++)
+    for (int c = 0; c < 0x1000; c++)
     {
 	if (BlockIsROM [c])
 	    WriteMap [c] = (uint8 *) MAP_NONE;
@@ -673,22 +773,22 @@ void CMemory::MapRAM ()
     int c;
 
     // Banks 7e->7f, RAM
-    for (c = 0; c < 8; c++)
+    for (c = 0; c < 16; c++)
     {
-	Map [c + 0x3f0] = RAM;
-	Map [c + 0x3f8] = RAM + 0x10000;
-	BlockIsRAM [c + 0x3f0] = TRUE;
-	BlockIsRAM [c + 0x3f8] = TRUE;
-	BlockIsROM [c + 0x3f0] = FALSE;
-	BlockIsROM [c + 0x3f8] = FALSE;
+	Map [c + 0x7e0] = RAM;
+	Map [c + 0x7f0] = RAM + 0x10000;
+	BlockIsRAM [c + 0x7e0] = TRUE;
+	BlockIsRAM [c + 0x7f0] = TRUE;
+	BlockIsROM [c + 0x7e0] = FALSE;
+	BlockIsROM [c + 0x7f0] = FALSE;
     }
 
     // Banks 70->77, S-RAM
-    for (c = 0; c < 0x40; c++)
+    for (c = 0; c < 0x80; c++)
     {
-	Map [c + 0x380] = (uint8 *) MAP_LOROM_SRAM;
-	BlockIsRAM [c + 0x380] = TRUE;
-	BlockIsROM [c + 0x380] = FALSE;
+	Map [c + 0x700] = (uint8 *) MAP_LOROM_SRAM;
+	BlockIsRAM [c + 0x700] = TRUE;
+	BlockIsROM [c + 0x700] = FALSE;
     }
 }
 
@@ -698,70 +798,78 @@ void CMemory::LoROMMap ()
     int i;
 
     // Banks 00->3f and 80->bf
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
 	if (Settings.DSP1Master)
-	    Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_DSP;
-	else
-	    Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_NONE;
-	for (i = c + 4; i < c + 8; i++)
 	{
-	    Map [i] = Map [i + 0x400] = &ROM [(c >> 1) << 13] - 0x8000;
-	    BlockIsROM [i] = BlockIsROM [i + 0x400] = TRUE;
+	    Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_DSP;
+	    Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_DSP;
+	}
+	else
+	{
+	    Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_NONE;
+	    Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_NONE;
+	}
+	for (i = c + 8; i < c + 16; i++)
+	{
+	    Map [i] = Map [i + 0x800] = &ROM [c << 11] - 0x8000;
+	    BlockIsROM [i] = BlockIsROM [i + 0x800] = TRUE;
 	}
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
 
     if (Settings.DSP1Master)
     {
 	// Banks 30->3f and b0->bf
-	for (c = 0x180; c < 0x200; c += 8)
+	for (c = 0x300; c < 0x400; c += 16)
 	{
-	    Map [c + 4] = Map [c + 0x404] = (uint8 *) MAP_DSP;
-	    Map [c + 5] = Map [c + 0x405] = (uint8 *) MAP_DSP;
-	    Map [c + 6] = Map [c + 0x406] = (uint8 *) MAP_DSP;
-	    Map [c + 7] = Map [c + 0x407] = (uint8 *) MAP_DSP;
-	    BlockIsROM [c + 4] = BlockIsROM [c + 0x404] = FALSE;
-	    BlockIsROM [c + 5] = BlockIsROM [c + 0x405] = FALSE;
-	    BlockIsROM [c + 6] = BlockIsROM [c + 0x406] = FALSE;
-	    BlockIsROM [c + 7] = BlockIsROM [c + 0x407] = FALSE;
+	    for (i = c + 8; i < c + 16; i++)
+	    {
+		Map [i] = Map [i + 0x800] = (uint8 *) MAP_DSP;
+		BlockIsROM [i] = BlockIsROM [i + 0x800] = FALSE;
+	    }
 	}
     }
 
     // Banks 40->7f and c0->ff
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 4; i++)
-	    Map [i + 0x200] = Map [i + 0x600] = &ROM [((c >> 1) << 13) + 0x200000];
-
-	for (i = c + 4; i < c + 8; i++)
-	    Map [i + 0x200] = Map [i + 0x600] = &ROM [((c >> 1) << 13) + 0x200000 - 0x8000];
-
 	for (i = c; i < c + 8; i++)
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [(c << 11) + 0x200000];
+
+	for (i = c + 8; i < c + 16; i++)
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [(c << 11) + 0x200000 - 0x8000];
+
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    BlockIsROM [i + 0x200] = BlockIsROM [i + 0x600] = TRUE;
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0x400] = BlockIsROM [i + 0xc00] = TRUE;
 	}
     }
 
     if (Settings.DSP1Master)
     {
-	for (c = 0; c < 0x80; c++)
+	for (c = 0; c < 0x100; c++)
 	{
-	    Map [c + 0x700] = (uint8 *) MAP_DSP;
-	    MemorySpeed [c + 0x700] = 8;
-	    BlockIsROM [c + 0x700] = FALSE;
+	    Map [c + 0xe00] = (uint8 *) MAP_DSP;
+	    MemorySpeed [c + 0xe00] = 8;
+	    BlockIsROM [c + 0xe00] = FALSE;
 	}
     }
     MapRAM ();
@@ -774,49 +882,64 @@ void CMemory::HiROMMap ()
     int i;
 
     // Banks 00->3f and 80->bf
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
 	if (Settings.DSP1Master)
-	    Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_DSP;
-	else
-	    Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_NONE;
-	    
-	for (i = c + 4; i < c + 8; i++)
 	{
-	    Map [i] = Map [i + 0x400] = &ROM [c << 13];
-	    BlockIsROM [i] = BlockIsROM [i + 0x400] = TRUE;
+	    Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_DSP;
+	    Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_DSP;
+	}
+	else
+	{
+	    Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_NONE;
+	    Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_NONE;
+	}
+	    
+	for (i = c + 8; i < c + 16; i++)
+	{
+	    Map [i] = Map [i + 0x800] = &ROM [c << 12];
+	    BlockIsROM [i] = BlockIsROM [i + 0x800] = TRUE;
 	}
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
+	    int ppu = i & 15;
+	    
 	    MemorySpeed [i] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
 
-    // Banks 30->3f address ranges 6000->7ffff is S-RAM.
+    // Banks 30->3f and b0->bf, address ranges 6000->7fff is S-RAM.
     for (c = 0; c < 16; c++)
     {
-	Map [0x183 + (c << 3)] = (uint8 *) MAP_HIROM_SRAM;
-	Map [0x583 + (c << 3)] = (uint8 *) MAP_HIROM_SRAM;
-	BlockIsRAM [0x183 + (c << 3)] = TRUE;
-	BlockIsRAM [0x583 + (c << 3)] = TRUE;
+	Map [0x306 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0x307 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0xb06 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0xb07 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	BlockIsRAM [0x306 + (c << 4)] = TRUE;
+	BlockIsRAM [0x307 + (c << 4)] = TRUE;
+	BlockIsRAM [0xb06 + (c << 4)] = TRUE;
+	BlockIsRAM [0xb07 + (c << 4)] = TRUE;
     }
 
     // Banks 40->7f and c0->ff
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    Map [i + 0x200] = Map [i + 0x600] = &ROM [c << 13];
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    BlockIsROM [i + 0x200] = BlockIsROM [i + 0x600] = TRUE;
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [c << 12];
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0x400] = BlockIsROM [i + 0xc00] = TRUE;
 	}
     }
 
@@ -830,35 +953,47 @@ void CMemory::TalesROMMap ()
     int i;
 
     // Banks 00->3f and 80->bf
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
-	Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_NONE;
-	for (i = c + 4; i < c + 8; i++)
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
+	Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_NONE;
+	Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_NONE;
+	for (i = c + 8; i < c + 16; i++)
 	{
-	    Map [i] = &ROM [c << 13];
-	    Map [i + 0x400] = &ROM [c << 13];
+	    Map [i] = &ROM [c << 12];
+	    Map [i + 0x800] = &ROM [c << 12];
 	    BlockIsROM [i] = TRUE;
-	    BlockIsROM [i + 0x400] = TRUE;
+	    BlockIsROM [i + 0x800] = TRUE;
 	}
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
     
-    // Banks 30->3f address ranges 6000->7ffff is S-RAM.
+    // Banks 30->3f and b0->bf, address ranges 6000->7ffff is S-RAM.
     for (c = 0; c < 16; c++)
     {
-	Map [0x583 + (c << 3)] = (uint8 *) MAP_HIROM_SRAM;
-	BlockIsRAM [0x583 + (c << 3)] = TRUE;
+	Map [0x306 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0x307 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0xb06 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	Map [0xb07 + (c << 4)] = (uint8 *) MAP_HIROM_SRAM;
+	BlockIsRAM [0x306 + (c << 4)] = TRUE;
+	BlockIsRAM [0x307 + (c << 4)] = TRUE;
+	BlockIsRAM [0xb06 + (c << 4)] = TRUE;
+	BlockIsRAM [0xb07 + (c << 4)] = TRUE;
     }
 
 #define OFFSET0 0x100000
@@ -867,20 +1002,20 @@ void CMemory::TalesROMMap ()
 #define OFFSET3 0x200000
 
     // Banks 40->7f and c0->ff
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 4; i++)
+	for (i = c; i < c + 8; i++)
 	{
-	    Map [i + 0x200] = &ROM [c << 13];
-	    Map [i + 0x204] = &ROM [c << 13];
-	    Map [i + 0x600] = &ROM [c << 13] + OFFSET3;
-	    Map [i + 0x604] = &ROM [c << 13] + OFFSET3;
-	    BlockIsROM [i + 0x200] = TRUE;
-	    BlockIsROM [i + 0x204] = TRUE;
-	    BlockIsROM [i + 0x600] = TRUE;
-	    BlockIsROM [i + 0x604] = TRUE;
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    MemorySpeed [i + 0x204] = MemorySpeed [i + 0x604] = 8;
+	    Map [i + 0x400] = &ROM [c << 12];
+	    Map [i + 0x408] = &ROM [c << 12];
+	    Map [i + 0xc00] = &ROM [c << 12] + OFFSET3;
+	    Map [i + 0xc08] = &ROM [c << 12] + OFFSET3;
+	    BlockIsROM [i + 0x400] = TRUE;
+	    BlockIsROM [i + 0x408] = TRUE;
+	    BlockIsROM [i + 0xc00] = TRUE;
+	    BlockIsROM [i + 0xc08] = TRUE;
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    MemorySpeed [i + 0x408] = MemorySpeed [i + 0xc08] = 8;
 	}
     }
     MapRAM ();
@@ -894,80 +1029,50 @@ void CMemory::AlphaROMMap ()
     int i;
 
     // Banks 00->3f and 80->bf
-#if 0
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
-	Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_DSP;
-	for (i = c + 4; i < c + 8; i++)
-	{
-	    Map [i] = Map [i + 0x400] = &ROM [c << 13];
-	    BlockIsROM [i] = BlockIsROM [i + 0x400] = TRUE;
-	}
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
+	Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_DSP;
+	Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_DSP;
 
-	for (i = c; i < c + 8; i++)
+	for (i = c + 8; i < c + 16; i++)
 	{
-	    MemorySpeed [i] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
-	}
-    }
-#else
-    for (c = 0; c < 0x200; c += 8)
-    {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
-
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
-	Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_DSP;
-	for (i = c + 4; i < c + 8; i++)
-	{
-	    Map [i] = Map [i + 0x400] = &ROM [(c >> 1) << 13] - 0x8000;
+	    Map [i] = Map [i + 0x800] = &ROM [c << 11] - 0x8000;
 	    BlockIsROM [i] = TRUE;
 	}
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
-#endif    
+
     // Banks 40->7f and c0->ff
 
 #undef OFFSET
 #define OFFSET 0x000000
 
-#if 0
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 4; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    Map [i + 0x200] = Map [i + 0x600] = &ROM [((c >> 1) << 13) + 0x200000];
-	    Map [i + 0x204] = Map [i + 0x604] = &ROM [((c >> 1) << 13) + 0x200000 - 0x8000];
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    MemorySpeed [i + 0x204] = MemorySpeed [i + 0x604] = 8;
-	    BlockIsROM [i + 0x200] = BlockIsROM [i + 0x204] = TRUE;
-	    BlockIsROM [i + 0x600] = BlockIsROM [i + 0x604] = TRUE;
-    }
-#else
-    for (c = 0; c < 0x200; c += 8)
-    {
-	for (i = c; i < c + 8; i++)
-	{
-	    Map [i + 0x200] = (uint8 *) MAP_DEBUG;
-	    Map [i + 0x600] = &ROM [c << 13] + OFFSET;
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    BlockIsROM [i + 0x200] = BlockIsROM [i + 0x600] = TRUE;
+	    Map [i + 0x400] = (uint8 *) MAP_DEBUG;
+	    Map [i + 0xc00] = &ROM [c << 12] + OFFSET;
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0x400] = BlockIsROM [i + 0xc00] = TRUE;
 	}
     }
-#endif
 
     MapRAM ();
     WriteProtectROM ();
@@ -977,68 +1082,84 @@ void CMemory::SuperFXROMMap ()
 {
     int c;
     int i;
-    int num_blocks = (1024 * 1024 * 2) / 0x2000;
     
     // Banks 00->3f and 80->bf
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
-	Map [c + 3] = Map [c + 0x403] = (uint8 *) MAP_DSP;
-	for (i = c + 4; i < c + 8; i++)
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
+	Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_DSP;
+	Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_DSP;
+	for (i = c + 8; i < c + 16; i++)
 	{
-	    Map [i] = Map [i + 0x400] = &ROM [(c >> 1) << 13] - 0x8000;
-	    BlockIsROM [i] = BlockIsROM [i + 0x400] = TRUE;
+	    Map [i] = Map [i + 0x800] = &ROM [c << 11] - 0x8000;
+	    BlockIsROM [i] = BlockIsROM [i + 0x800] = TRUE;
 	}
 
 	for (i = c; i < c + 8; i++)
 	{
-	    MemorySpeed [i + 0] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
     
     // Banks 40->7f and c0->ff
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    Map [i + 0x200] = Map [i + 0x600] = &ROM [c << 13];
-	    MemorySpeed [i + 0x200] = MemorySpeed [i + 0x600] = 8;
-	    BlockIsROM [i + 0x200] = BlockIsROM [i + 0x600] = TRUE;
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [c << 12];
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0x400] = BlockIsROM [i + 0xc00] = TRUE;
 	}
     }
 
     // Banks 7e->7f, RAM
-    for (c = 0; c < 8; c++)
+    for (c = 0; c < 16; c++)
     {
-	Map [c + 0x3f0] = RAM;
-	Map [c + 0x3f8] = RAM + 0x10000;
-	BlockIsRAM [c + 0x3f0] = TRUE;
-	BlockIsRAM [c + 0x3f8] = TRUE;
-	BlockIsROM [c + 0x3f0] = FALSE;
-	BlockIsROM [c + 0x3f8] = FALSE;
+	Map [c + 0x7e0] = RAM;
+	Map [c + 0x7f0] = RAM + 0x10000;
+	BlockIsRAM [c + 0x7e0] = TRUE;
+	BlockIsRAM [c + 0x7f0] = TRUE;
+	BlockIsROM [c + 0x7e0] = FALSE;
+	BlockIsROM [c + 0x7f0] = FALSE;
     }
 
     // Banks 70->71, S-RAM
-    for (c = 0; c < 0x10; c++)
+    for (c = 0; c < 32; c++)
     {
-	Map [c + 0x380] = ::SRAM + (((c >> 3) & 1) << 16);
-	BlockIsRAM [c + 0x380] = TRUE;
-	BlockIsROM [c + 0x380] = FALSE;
+	Map [c + 0x700] = ::SRAM + (((c >> 4) & 1) << 16);
+	BlockIsRAM [c + 0x700] = TRUE;
+	BlockIsROM [c + 0x700] = FALSE;
     }
 
     // Banks 00->3f and 80->bf address ranges 6000->7fff is RAM.
     for (c = 0; c < 0x40; c++)
     {
-	Map [0x003 + (c << 3)] = (uint8 *) ::SRAM - 0x6000;
-	Map [0x403 + (c << 3)] = (uint8 *) ::SRAM - 0x6000;
-	BlockIsRAM [0x003 + (c << 3)] = TRUE;
-	BlockIsRAM [0x403 + (c << 3)] = TRUE;
+	Map [0x006 + (c << 4)] = (uint8 *) ::SRAM - 0x6000;
+	Map [0x007 + (c << 4)] = (uint8 *) ::SRAM - 0x6000;
+	Map [0x806 + (c << 4)] = (uint8 *) ::SRAM - 0x6000;
+	Map [0x807 + (c << 4)] = (uint8 *) ::SRAM - 0x6000;
+	BlockIsRAM [0x006 + (c << 4)] = TRUE;
+	BlockIsRAM [0x007 + (c << 4)] = TRUE;
+	BlockIsRAM [0x806 + (c << 4)] = TRUE;
+	BlockIsRAM [0x807 + (c << 4)] = TRUE;
+    }
+    // Replicate the first 2Mb of the ROM at ROM + 2MB such that each 32K
+    // block is repeated twice in each 64K block.
+    for (c = 0; c < 64; c++)
+    {
+	memmove (&ROM [0x200000 + c * 0x10000], &ROM [c * 0x8000], 0x8000);
+	memmove (&ROM [0x208000 + c * 0x10000], &ROM [c * 0x8000], 0x8000);
     }
 
     WriteProtectROM ();
@@ -1050,67 +1171,174 @@ void CMemory::SA1ROMMap ()
     int i;
 
     // Banks 00->3f and 80->bf
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	Map [c + 0] = Map [c + 0x400] = RAM;
-	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x400] = TRUE;
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
 
-	Map [c + 1] = Map [c + 0x401] = (uint8 *) MAP_PPU;
-	Map [c + 2] = Map [c + 0x402] = (uint8 *) MAP_CPU;
-	Map [c + 3] = Map [c + 0x403] = (uint8 *) SRAM;
-	for (i = c + 4; i < c + 8; i++)
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) &Memory.FillRAM [0x3000] - 0x3000;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
+	Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_BWRAM;
+	Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_BWRAM;
+	for (i = c + 8; i < c + 16; i++)
 	{
-	    Map [i] = Map [i + 0x400] = &ROM [(c >> 1) << 13] - 0x8000;
-	    BlockIsROM [i] = BlockIsROM [i + 0x400] = TRUE;
+	    Map [i] = Map [i + 0x800] = &ROM [c << 11] - 0x8000;
+	    BlockIsROM [i] = BlockIsROM [i + 0x800] = TRUE;
 	}
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0] = 
-		MemorySpeed [i + 0x400] = (i & 7) == 1 || (i & 7) == 2 ? 
-					    ONE_CYCLE : 8;
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
 	}
     }
 
     // Banks 40->7f
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 8; i++)
-	    Map [i + 0x200] = &SRAM [(c << 13) & 0xffff];
+	for (i = c; i < c + 16; i++)
+	    Map [i + 0x400] = (uint8 *) &SRAM [(c << 12) & 0xffff];
 
-	for (i = c; i < c + 8; i++)
+	for (i = c; i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0x200] = 8;
-	    BlockIsROM [i + 0x200] = TRUE;
+	    MemorySpeed [i + 0x400] = 8;
+	    BlockIsROM [i + 0x400] = FALSE;
 	}
     }
 
     // c0->ff
-    for (c = 0; c < 0x200; c += 8)
+    for (c = 0; c < 0x400; c += 16)
     {
-	for (i = c; i < c + 4; i++)
-	    Map [i + 0x600] = &ROM [c << 13];
-
-	for (i = c + 4; i < c + 8; i++)
-	    Map [i + 0x600] = &ROM [c << 13];
-
-	for (i = c; i < c + 8; i++)
+	for (i = c;  i < c + 16; i++)
 	{
-	    MemorySpeed [i + 0x600] = 8;
-	    BlockIsROM [i + 0x600] = TRUE;
+	    Map [i + 0xc00] = &ROM [c << 12];
+	    MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0xc00] = TRUE;
 	}
     }
 
-    for (c = 0; c < 8; c++)
+    for (c = 0; c < 16; c++)
     {
-	Map [c + 0x3f0] = RAM;
-	Map [c + 0x3f8] = RAM + 0x10000;
-	BlockIsRAM [c + 0x3f0] = TRUE;
-	BlockIsRAM [c + 0x3f8] = TRUE;
-	BlockIsROM [c + 0x3f0] = FALSE;
-	BlockIsROM [c + 0x3f8] = FALSE;
+	Map [c + 0x7e0] = RAM;
+	Map [c + 0x7f0] = RAM + 0x10000;
+	BlockIsRAM [c + 0x7e0] = TRUE;
+	BlockIsRAM [c + 0x7f0] = TRUE;
+	BlockIsROM [c + 0x7e0] = FALSE;
+	BlockIsROM [c + 0x7f0] = FALSE;
     }
     WriteProtectROM ();
+
+    // Now copy the map and correct it for the SA1 CPU.
+    memmove ((void *) SA1.WriteMap, (void *) WriteMap, sizeof (WriteMap));
+    memmove ((void *) SA1.Map, (void *) Map, sizeof (Map));
+    memmove ((void *) SA1.BlockIsRAM, (void *) BlockIsRAM, sizeof (BlockIsRAM));
+
+    // Banks 00->3f and 80->bf
+    for (c = 0; c < 0x400; c += 16)
+    {
+	SA1.Map [c + 0] = SA1.Map [c + 0x800] = &Memory.FillRAM [0x3000];
+	SA1.Map [c + 1] = SA1.Map [c + 0x801] = (uint8 *) MAP_NONE;
+	SA1.WriteMap [c + 0] = SA1.WriteMap [c + 0x800] = &Memory.FillRAM [0x3000];
+	SA1.WriteMap [c + 1] = SA1.WriteMap [c + 0x801] = (uint8 *) MAP_NONE;
+	SA1.BlockIsRAM [c + 0] = SA1.BlockIsRAM [c + 0x800] = FALSE;
+	SA1.BlockIsRAM [c + 1] = SA1.BlockIsRAM [c + 0x801] = FALSE;
+    }
+
+    // Banks 60->6f
+    for (c = 0; c < 0x100; c++)
+	SA1.Map [c + 0x600] = SA1.WriteMap [c + 0x600] = (uint8 *) MAP_BWRAM_BITMAP;
+    
+    BWRAM = SRAM;
+}
+
+void CMemory::LoROM24MBSMap ()
+{
+    int c;
+    int i;
+
+    // Banks 00->3f and 80->bf
+    for (c = 0; c < 0x400; c += 16)
+    {
+	Map [c + 0] = Map [c + 0x800] = RAM;
+	Map [c + 1] = Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0] = BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 1] = BlockIsRAM [c + 0x801] = TRUE;
+
+	Map [c + 2] = Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 3] = Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 4] = Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 5] = Map [c + 0x805] = (uint8 *) MAP_CPU;
+       Map [c + 6] = Map [c + 0x806] = (uint8 *) MAP_NONE;
+       Map [c + 7] = Map [c + 0x807] = (uint8 *) MAP_NONE;
+
+	for (i = c + 8; i < c + 16; i++)
+	{
+	    Map [i] = Map [i + 0x800] = &ROM [c << 11] - 0x8000;
+	    BlockIsROM [i] = BlockIsROM [i + 0x800] = TRUE;
+	}
+
+	for (i = c; i < c + 16; i++)
+	{
+	    int ppu = i & 15;
+	    
+	    MemorySpeed [i] = 
+		MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
+	}
+    }
+
+    // Banks 00->3f and 80->bf
+    for (c = 0; c < 0x200; c += 16)
+    {
+	Map [c + 0x800] = RAM;
+	Map [c + 0x801] = RAM;
+	BlockIsRAM [c + 0x800] = TRUE;
+	BlockIsRAM [c + 0x801] = TRUE;
+
+	Map [c + 0x802] = (uint8 *) MAP_PPU;
+	Map [c + 0x803] = (uint8 *) MAP_PPU;
+	Map [c + 0x804] = (uint8 *) MAP_CPU;
+	Map [c + 0x805] = (uint8 *) MAP_CPU;
+       Map [c + 0x806] = (uint8 *) MAP_NONE;
+	Map [c + 0x807] = (uint8 *) MAP_NONE;
+
+	for (i = c + 8; i < c + 16; i++)
+	{
+	    Map [i + 0x800] = &ROM [c << 11] - 0x8000 + 0x200000;
+	    BlockIsROM [i + 0x800] = TRUE;
+	}
+
+	for (i = c; i < c + 16; i++)
+	{
+	    int ppu = i & 15;
+	    
+	MemorySpeed [i + 0x800] = ppu >= 2 && ppu <= 3 ? ONE_CYCLE : 8;
+	}
+    }
+
+    // Banks 40->7f and c0->ff
+    for (c = 0; c < 0x400; c += 16)
+    {
+	for (i = c; i < c + 8; i++)
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [(c << 11) + 0x200000];
+
+	for (i = c + 8; i < c + 16; i++)
+	    Map [i + 0x400] = Map [i + 0xc00] = &ROM [(c << 11) + 0x200000 - 0x8000];
+
+	for (i = c; i < c + 16; i++)
+	{
+	    MemorySpeed [i + 0x400] = MemorySpeed [i + 0xc00] = 8;
+	    BlockIsROM [i + 0x400] = BlockIsROM [i + 0xc00] = TRUE;
+	}
+    }
+
+    MapRAM ();
+    WriteProtectROM ();
 }
 
 const char *CMemory::TVStandard ()
@@ -1192,6 +1420,35 @@ void CMemory::ApplyROMFixes ()
 
     // XXX: Battle Toads
     Settings.BattleToadsNMIHack = FALSE;
+
+    if (strcmp (ROMName, "TETRIS&Dr.MARIO") == 0 ||
+	strcmp (ROMName, "JIGSAW PARTY") == 0)
+    {
+	Settings.MultiPlayer5Master = FALSE;
+	Settings.MouseMaster = FALSE;
+	Settings.SuperScopeMaster = FALSE;
+    }
+
+    // Games written by the Japanese software company called Human need special
+    // SPC700 sound CPU cycle length timing to get them to work...
+    if ((Memory.HiROM && strncmp ((char *) &Memory.ROM [0xffb0], "DE", 2) == 0) ||
+	(!Memory.HiROM && strncmp ((char *) &Memory.ROM [0x7fb0], "DE", 2) == 0) ||
+	strcmp (ROMName, "SUPER FORMATION SOCCE") == 0 ||
+	strcmp (ROMName, "HOOK") == 0 ||
+	strcmp (ROMName, "SUPER EMPIRE STRIKES") == 0)
+    {
+	IAPU.OneCycle = ONE_APU_CYCLE_HUMAN;
+    }
+    if (strcmp (ROMName, "EARTHWORM JIM 2") == 0 ||
+	strcmp (ROMName, "ULTIMATE KOMBAT 3") == 0 ||
+	strncmp (ROMName, "MORTAL KOMBAT 3", 15) == 0)
+	IAPU.OneCycle = 30;
+
+    Settings.DaffyDuckNMIHack = strcmp (ROMName, "DAFFY DUCK: MARV MISS") == 0;
+    Settings.StarfoxHack = strcmp (ROMName, "STAR FOX") == 0 ||
+			   strcmp (ROMName, "STAR WING") == 0;
+    Settings.WinterGold = strcmp (ROMName, "FX SKIING NINTENDO 96") == 0;
+    Settings.Seiken3 = strncmp (ROMName, "SeikenDensetsu3", 15) == 0;  
 }
 
 #ifdef NO_INLINE_SET_GET
