@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -313,10 +313,14 @@ void S9xInitDisplay (int argc, char **argv)
 	    S9xExit ();
 	}
 	GFX.SubScreen = (uint8 *) sub_screen->dat;
+	GFX.ZBuffer = (uint8 *) malloc (off_screen->w * IMAGE_HEIGHT);
+	GFX.SubZBuffer = (uint8 *) malloc (off_screen->w * IMAGE_HEIGHT);
     }
     else
     {
 	GFX.SubScreen = GFX.Screen;
+	GFX.ZBuffer = (uint8 *) malloc (off_screen->w * IMAGE_HEIGHT);
+	GFX.SubZBuffer = NULL;
 	sub_screen = NULL;
 	Delta = NULL;
     }
@@ -334,8 +338,16 @@ void S9xDeinitDisplay ()
     }
 #endif
     destroy_bitmap (off_screen);
-    destroy_bitmap (sub_screen);
+    if (sub_screen)
+	destroy_bitmap (sub_screen);
+    
     delete Delta;
+    if (GFX.ZBuffer)
+	free ((char *) GFX.ZBuffer);
+    if (GFX.SubZBuffer)
+	free ((char *) GFX.SubZBuffer);
+    GFX.ZBuffer = NULL;
+    GFX.SubZBuffer = NULL;
 }
 
 void S9xInitInputDevices ()
@@ -1134,14 +1146,20 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
     atexit (sx_sb_exit);
 
     /* Lock variables into memory */
+    
     void S9xMixSamples_end();
+    void S9xMixSamplesO_end();
     void MixStereo (int);
     void MixStereo_end();
     void MixMono (int);
     void MixMono_end();
+    void S9xAPUSetEndOfSample (int, Channel *);
     void S9xAPUSetEndOfSample_end();
+    void S9xAPUSetEndX (int);
     void S9xAPUSetEndX_end();
     void S9xSetEnvelopeRate_end();
+    void S9xSetEnvRate (Channel *, unsigned long, int, int);
+    void S9xSetEnvRate_end ();
     
     extern long FilterValues[4][2];
     extern int Echo [15360 + 14];
@@ -1153,11 +1171,13 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
     extern int Loop [16];
 
     LOCK_FUNCTION (S9xMixSamples);
+    LOCK_FUNCTION (S9xMixSamplesO);
     LOCK_FUNCTION (MixStereo);
     LOCK_FUNCTION (MixMono);
     LOCK_FUNCTION (S9xAPUSetEndX);
     LOCK_FUNCTION (S9xAPUSetEndOfSample);
     LOCK_FUNCTION (S9xSetEnvelopeRate);
+    LOCK_FUNCTION (S9xSetEnvRate);
     LOCK_VARIABLE (APU);
     LOCK_VARIABLE (IAPU);
     LOCK_VARIABLE (SoundData);
@@ -1183,6 +1203,10 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
     return (TRUE);
 }
 
+void S9xGenerateSound ()
+{
+}
+
 bool8 S9xReadMousePosition (int which1, int &x, int &y, uint32 &buttons)
 {
     if (which1 == 0)
