@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -37,7 +37,6 @@
  */
 
 #include "snes9x.h"
-
 #include "memmap.h"
 #include "ppu.h"
 #include "cpuexec.h"
@@ -46,6 +45,7 @@
 #include "dma.h"
 #include "gfx.h"
 #include "display.h"
+#include "sa1.h"
 
 #ifndef ZSNES_FX
 #include "fxemu.h"
@@ -90,6 +90,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	    if (PPU.Brightness != (Byte & 0xF))
 	    {
 		IPPU.ColorsChanged = TRUE;
+		IPPU.DirectColourMapsNeedRebuild = TRUE;
 		PPU.Brightness = Byte & 0xF;
 		S9xFixColourBrightness ();
 		if (PPU.Brightness > IPPU.MaxBrightness)
@@ -108,7 +109,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	if (Byte != Memory.FillRAM [0x2101])
 	{
 	    FLUSH_REDRAW ();
-	    PPU.OBJNameBase	= (Byte & 3) << 14;
+	    PPU.OBJNameBase   = (Byte & 3) << 14;
 	    PPU.OBJNameSelect = ((Byte >> 3) & 3) << 13;
 	    PPU.OBJSizeSelect = (Byte >> 5) & 7;
 	    IPPU.OBJChanged = TRUE;
@@ -117,8 +118,9 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 
     case 0x2102:
 	// Sprite write address (low)
-	PPU.OAMAddr &= 0xFF00;
-	PPU.OAMAddr |= Byte;
+	//PPU.OAMAddr &= 0xFF00;
+	//PPU.OAMAddr |= Byte;
+	PPU.OAMAddr = Byte;
 	PPU.OAMFlip = 0;
 	PPU.OAMReadFlip = 0;
 	PPU.SavedOAMAddr = PPU.OAMAddr;
@@ -143,8 +145,9 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	    missing.sprite_priority_rotation = 1;
 #endif
 	}
-	if (PPU.OAMAddr >= 0x110)
-	    PPU.OAMAddr %= 0x110;
+	// No more wrap around...
+	//if (PPU.OAMAddr >= 0x110)
+	//    PPU.OAMAddr %= 0x110;
 	PPU.OAMFlip = 0;
 	PPU.OAMReadFlip = 0;
 	PPU.SavedOAMAddr = PPU.OAMAddr;
@@ -161,11 +164,11 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	if (Byte != Memory.FillRAM [0x2105])
 	{
 	    FLUSH_REDRAW ();
-	    PPU.BG3Priority	= (Byte >> 3) & 1;
-	    PPU.BG[0].BGSize= (Byte >> 4) & 1;
-	    PPU.BG[1].BGSize= (Byte >> 5) & 1;
-	    PPU.BG[2].BGSize= (Byte >> 6) & 1;
-	    PPU.BG[3].BGSize= (Byte >> 7) & 1;
+	    PPU.BG3Priority  = (Byte >> 3) & 1;
+	    PPU.BG[0].BGSize = (Byte >> 4) & 1;
+	    PPU.BG[1].BGSize = (Byte >> 5) & 1;
+	    PPU.BG[2].BGSize = (Byte >> 6) & 1;
+	    PPU.BG[3].BGSize = (Byte >> 7) & 1;
 	    PPU.BGMode = Byte & 7;
 #ifdef DEBUGGER
 	    missing.modes[PPU.BGMode] = 1;
@@ -575,6 +578,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	if (Byte != Memory.FillRAM [0x2130])
 	{
 	    FLUSH_REDRAW ();
+	    PPU.RecomputeClipWindows = TRUE;
 #ifdef DEBUGGER
 	    if ((Byte & 1) && (PPU.BGMode == 3 || PPU.BGMode == 4 || PPU.BGMode == 7))
 		missing.direct = 1;
@@ -703,7 +707,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 #ifdef SPC700_SHUTDOWN	
 	IAPU.APUExecuting = Settings.APUEnabled;
 	IAPU.WaitCounter++;
-#endif	
+#endif
 	break;
     case 0x2180:
 	REGISTER_2180(Byte);
@@ -725,6 +729,15 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
     }
     else
     {
+	if (Settings.SA1)
+	{
+	    if (Address >= 0x2200 && Address <0x23ff)
+		S9xSetSA1 (Byte, Address);
+	    else
+		Memory.FillRAM [Address] = Byte;
+	    return;
+	}
+	else
     if (Address < 0x3000 || Address >= 0x3000 + 768)
     {
 #ifdef DEBUGGER
@@ -762,6 +775,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	    else
 		Memory.FillRAM [Address] = Byte;
 	    break;
+
 	case 0x3031:
 	    Memory.FillRAM [Address] = Byte;
 	    break;
@@ -826,15 +840,15 @@ uint8 S9xGetPPU (uint16 Address)
 	{
 	case 0x2100:
 	case 0x2101:
-	case 0x2102:
 	    return (Memory.FillRAM[Address]);
-	case 0x2103:
+	case 0x2102:
 #ifdef DEBUGGER
 	    missing.oam_address_read = 1;
 #endif
 	    return (uint8)(PPU.OAMAddr);
-	case 0x2104:
+	case 0x2103:
 	    return (((PPU.OAMAddr >> 8) & 1) | (PPU.OAMPriorityRotation << 7));
+	case 0x2104:
 	case 0x2105:
 	case 0x2106:
 	case 0x2107:
@@ -919,6 +933,11 @@ uint8 S9xGetPPU (uint16 Address)
 	    // Latch h and v counters
 #ifdef DEBUGGER
 	    missing.h_v_latch = 1;
+#endif
+#if 0
+#ifdef CPU_SHUTDOWN
+	    CPU.WaitAddress = CPU.PCAtOpcodeStart;
+#endif
 #endif
 	    PPU.HVBeamCounterLatched = 1;
 	    PPU.VBeamPosLatched = CPU.V_Counter;
@@ -1057,8 +1076,17 @@ uint8 S9xGetPPU (uint16 Address)
 	case 0x2178: case 0x2179: case 0x217a: case 0x217b:
 	case 0x217c: case 0x217d: case 0x217e: case 0x217f:
     //	CPU.Flags |= DEBUG_MODE_FLAG;
+#ifdef SPC700_SHUTDOWN	
+	IAPU.APUExecuting = Settings.APUEnabled;
+	IAPU.WaitCounter++;
+#endif
 	    if (Settings.APUEnabled)
+	    {
+#ifdef CPU_SHUTDOWN
+//		CPU.WaitAddress = CPU.PCAtOpcodeStart;
+#endif	
 		return (APU.OutPorts [Address & 3]);
+	    }
 
 	    switch (Settings.SoundSkipMethod)
 	    {
@@ -1107,6 +1135,9 @@ uint8 S9xGetPPU (uint16 Address)
     }
     else
     {
+	if (Settings.SA1)
+	    return (S9xGetSA1 (Address));
+
 	if (Address <= 0x2fff || Address >= 0x3000 + 768)
 	{
 	    switch (Address)
@@ -1135,6 +1166,7 @@ uint8 S9xGetPPU (uint16 Address)
 	    Byte = S9xSuperFXReadReg (Address);
 	else
 	    Byte = Memory.FillRAM [Address];
+
 #ifdef CPU_SHUTDOWN
 	if (Address == 0x3030)
 	    CPU.WaitAddress = CPU.PCAtOpcodeStart;
@@ -1144,6 +1176,8 @@ uint8 S9xGetPPU (uint16 Address)
 #else
 	Byte = Memory.FillRAM [Address];
 
+//if (Address != 0x3030 && Address != 0x3031)
+//printf ("%04x\n", Address);
 #ifdef CPU_SHUTDOWN
 	if (Address == 0x3030)
 	{
@@ -1182,7 +1216,8 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	    // S9xReset reading of old-style joypads
 	    if ((byte & 1) && !(Memory.FillRAM [Address] & 1))
 	    {
-		PPU.Joypad1ButtonReadPos = PPU.Joypad2ButtonReadPos = 0;
+		PPU.Joypad1ButtonReadPos = 0;
+		PPU.Joypad2ButtonReadPos = 0;
 		PPU.Joypad3ButtonReadPos = 0;
 	    }
 	    break;
@@ -1236,6 +1271,17 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	    PPU.HTimerPosition = Settings.H_Max + 1;
 	}
 	CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
+
+	// Needed for Daffy Duck: The Marvin Missions.
+	// ...but causes problems for Breatch of Fire 1 and 2
+	if (Settings.DaffyDuckNMIHack && (byte & 0x80) && 
+	    !(Memory.FillRAM [0x4200] & 0x80) &&
+	    CPU.V_Counter == PPU.ScreenHeight + FIRST_VISIBLE_LINE)
+	{
+	    CPU.Flags &= ~DELAYED_NMI_FLAG;
+	    CPU.Flags |= NMI_FLAG;
+	    CPU.NMIActive = TRUE;
+	}
 	break;
     case 0x4201:
 	// I/O port output 
@@ -1272,6 +1318,9 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	PPU.IRQHBeamPos = (PPU.IRQHBeamPos & 0xFF00) | byte;
 	if (PPU.HTimerEnabled)
 	{
+#ifdef DEBUGGER
+	    missing.hirq_pos = PPU.IRQHBeamPos;
+#endif
 	    PPU.HTimerPosition = PPU.IRQHBeamPos * Settings.H_Max / SNES_HCOUNTER_MAX;
 	    if (PPU.HTimerPosition == Settings.H_Max ||
 		PPU.HTimerPosition == Settings.HBlankStart)
@@ -1286,11 +1335,14 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 		{
 		    CPU.NextEvent = PPU.HTimerPosition;
 		    CPU.WhichEvent = HTIMER_EVENT;
+		    break;
 		}
-	    }
-#ifdef DEBUGGER
-	    missing.hirq_pos = PPU.IRQHBeamPos;
+#if 0
+// Disabled because it causes problems for Ninja Warriors
+		CPU.Flags |= IRQ_PENDING_FLAG;
+		CPU.IRQActive |= PPU_H_BEAM_IRQ_SOURCE;
 #endif
+	    }
 	}
 	break;
     case 0x4208:
@@ -1298,6 +1350,9 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	    ((byte & 1) << 8);
 	if (PPU.HTimerEnabled)
 	{
+#ifdef DEBUGGER
+	    missing.hirq_pos = PPU.IRQHBeamPos;
+#endif
 	    PPU.HTimerPosition = (PPU.IRQHBeamPos * Settings.H_Max) / SNES_HCOUNTER_MAX;
 	    if (PPU.HTimerPosition == Settings.H_Max ||
 		PPU.HTimerPosition == Settings.HBlankStart)
@@ -1312,45 +1367,64 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 		{
 		    CPU.NextEvent = PPU.HTimerPosition;
 		    CPU.WhichEvent = HTIMER_EVENT;
+		    break;
 		}
-	    }
-#ifdef DEBUGGER
-	    missing.hirq_pos = PPU.IRQHBeamPos;
+#if 0
+// Disabled because it causes problems for Ninja Warriors
+		CPU.Flags |= IRQ_PENDING_FLAG;
+		CPU.IRQActive |= PPU_H_BEAM_IRQ_SOURCE;
 #endif
+	    }
 	}
 	break;
     case 0x4209:
 	PPU.IRQVBeamPos = (PPU.IRQVBeamPos & 0xFF00) | byte;
-	if (PPU.VTimerEnabled && PPU.HTimerEnabled &&
-	    CPU.V_Counter == PPU.IRQVBeamPos)
-	{
-	    if (PPU.HTimerPosition >= CPU.Cycles &&
-		PPU.HTimerPosition < CPU.NextEvent)
-	    {
-		CPU.NextEvent = PPU.HTimerPosition;
-		CPU.WhichEvent = HTIMER_EVENT;
-	    }
-	}
 #ifdef DEBUGGER
 	missing.virq_pos = PPU.IRQVBeamPos;
 #endif
+	if (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos)
+	{
+	    if (PPU.HTimerEnabled)
+	    {
+		if (PPU.HTimerPosition >= CPU.Cycles &&
+		    PPU.HTimerPosition < CPU.NextEvent)
+		{
+		    CPU.NextEvent = PPU.HTimerPosition;
+		    CPU.WhichEvent = HTIMER_EVENT;
+		    break;
+		}
+	    }
+#if 0
+// Disabled for Chuck Rock
+	    CPU.Flags |= IRQ_PENDING_FLAG;
+	    CPU.IRQActive |= PPU_V_BEAM_IRQ_SOURCE;
+#endif
+	}
 	break;
     case 0x420A:
 	PPU.IRQVBeamPos = (PPU.IRQVBeamPos & 0xFF) |
 	    ((byte & 1) << 8);
-	if (PPU.VTimerEnabled && PPU.HTimerEnabled &&
-	    CPU.V_Counter == PPU.IRQVBeamPos)
-	{
-	    if (PPU.HTimerPosition >= CPU.Cycles &&
-		PPU.HTimerPosition < CPU.NextEvent)
-	    {
-		CPU.NextEvent = PPU.HTimerPosition;
-		CPU.WhichEvent = HTIMER_EVENT;
-	    }
-	}
 #ifdef DEBUGGER
 	missing.virq_pos = PPU.IRQVBeamPos;
 #endif
+	if (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos)
+	{
+	    if (PPU.HTimerEnabled)
+	    {
+		if (PPU.HTimerPosition >= CPU.Cycles &&
+		    PPU.HTimerPosition < CPU.NextEvent)
+		{
+		    CPU.NextEvent = PPU.HTimerPosition;
+		    CPU.WhichEvent = HTIMER_EVENT;
+		    break;
+		}
+	    }
+#if 0
+// Disabled for Chuck Rock
+	    CPU.Flags |= IRQ_PENDING_FLAG;
+	    CPU.IRQActive |= PPU_V_BEAM_IRQ_SOURCE;
+#endif
+	}
 	break;
     case 0x420B:
 #ifdef DEBUGGER
@@ -1673,7 +1747,6 @@ uint8 S9xGetCPU (uint16 Address)
 		}
 	    }
 	    return ((IPPU.Joypads[ind] >> (PPU.Joypad2ButtonReadPos++ ^ 15)) & 1);
-		    return (0x00);
         }
 	default:
 #ifdef DEBUGGER
@@ -1977,6 +2050,7 @@ void S9xResetPPU ()
     PPU.CentreX = PPU.CentreY = 0;
     PPU.Joypad1ButtonReadPos = 0;
     PPU.Joypad2ButtonReadPos = 0;
+    PPU.Joypad3ButtonReadPos = 0;
 
     PPU.CGADD = 0;
     PPU.FixedColourRed = PPU.FixedColourGreen = PPU.FixedColourBlue = 0;
@@ -2010,7 +2084,6 @@ void S9xResetPPU ()
     PPU.RecomputeClipWindows = TRUE;
     PPU.CGFLIPRead = 0;
     PPU.Need16x8Mulitply = FALSE;
-    PPU.Joypad3ButtonReadPos = 0;
     PPU.MouseSpeed[0] = PPU.MouseSpeed[1] = 0;
 
     IPPU.ColorsChanged = TRUE;
@@ -2020,6 +2093,7 @@ void S9xResetPPU ()
     IPPU.LatchedBlanking = 0;
     IPPU.OBJChanged = TRUE;
     IPPU.RenderThisFrame = TRUE;
+    IPPU.DirectColourMapsNeedRebuild = TRUE;
     IPPU.FrameCount = 0;
     IPPU.RenderedFramesCount = 0;
     IPPU.DisplayedRenderedFrameCount = 0;
@@ -2034,6 +2108,8 @@ void S9xResetPPU ()
     IPPU.RenderedScreenHeight = SNES_WIDTH;
     IPPU.RenderedScreenWidth = SNES_HEIGHT;
     IPPU.XB = NULL;
+    for (c = 0; c < 256; c++)
+	IPPU.ScreenColors [c] = c;
     S9xFixColourBrightness ();
     IPPU.PreviousLine = IPPU.CurrentLine = 0;
     IPPU.Joypads[0] = IPPU.Joypads[1] = IPPU.Joypads[2] = 0;
@@ -2218,42 +2294,58 @@ void S9xUpdateJoypads ()
 
     if (Memory.FillRAM [0x4200] & 1)
     {
-	PPU.Joypad1ButtonReadPos = PPU.Joypad2ButtonReadPos = 16;
-	PPU.Joypad3ButtonReadPos = 16;
-	Memory.FillRAM [0x4218] = Settings.SwapJoypads ? 
-					    (uint8) IPPU.Joypads [1] :
-					    (uint8) IPPU.Joypads [0];
-	Memory.FillRAM [0x4219] = Settings.SwapJoypads ?
-					    (uint8) (IPPU.Joypads [1] >> 8) :
-					    (uint8) (IPPU.Joypads [0] >> 8);
-	Memory.FillRAM [0x421a] = Settings.SwapJoypads ? 
-					    (uint8) IPPU.Joypads [0] :
-					    (uint8) IPPU.Joypads [1];
-	Memory.FillRAM [0x421b] = Settings.SwapJoypads ?
-					    (uint8) (IPPU.Joypads [0] >> 8) :
-					    (uint8) (IPPU.Joypads [1] >> 8);
-	Memory.FillRAM [0x421c] = (uint8) IPPU.Joypads [3];
-	Memory.FillRAM [0x421d] = (uint8) (IPPU.Joypads [3] >> 8);
-	Memory.FillRAM [0x421e] = (uint8) IPPU.Joypads [2];
-	Memory.FillRAM [0x421f] = (uint8) (IPPU.Joypads [2] >> 8);
+	PPU.Joypad1ButtonReadPos = 16;
+	if (Memory.FillRAM [0x4201] & 0x80)
+	{
+	    PPU.Joypad2ButtonReadPos = 16;
+	    PPU.Joypad3ButtonReadPos = 0;
+	}
+	else
+	{
+	    PPU.Joypad2ButtonReadPos = 0;
+	    PPU.Joypad3ButtonReadPos = 16;
+	}
+	int ind = Settings.SwapJoypads ? 1 : 0;
+
+	Memory.FillRAM [0x4218] = (uint8) IPPU.Joypads [ind];
+	Memory.FillRAM [0x4219] = (uint8) (IPPU.Joypads [ind] >> 8);
+	Memory.FillRAM [0x421a] = (uint8) IPPU.Joypads [ind ^ 1];
+	Memory.FillRAM [0x421b] = (uint8) (IPPU.Joypads [ind ^ 1] >> 8);
+	if (Memory.FillRAM [0x4201] & 0x80)
+	{
+	    Memory.FillRAM [0x421c] = (uint8) IPPU.Joypads [ind];
+	    Memory.FillRAM [0x421d] = (uint8) (IPPU.Joypads [ind] >> 8);
+	    Memory.FillRAM [0x421e] = (uint8) IPPU.Joypads [2];
+	    Memory.FillRAM [0x421f] = (uint8) (IPPU.Joypads [2] >> 8);
+	}
+	else
+	{
+	    Memory.FillRAM [0x421c] = (uint8) IPPU.Joypads [3];
+	    Memory.FillRAM [0x421d] = (uint8) (IPPU.Joypads [3] >> 8);
+	    Memory.FillRAM [0x421e] = (uint8) IPPU.Joypads [4];
+	    Memory.FillRAM [0x421f] = (uint8) (IPPU.Joypads [4] >> 8);
+	}
     }
 }
 
 #ifndef ZSNES_FX
 void S9xSuperFXExec ()
 {
-#ifdef EXECUTE_SUPERFX_PER_LINE
+#if 1
     if (Settings.SuperFX)
     {
 	if ((Memory.FillRAM [0x3000 + GSU_SFR] & FLG_G) &&
 	    (Memory.FillRAM [0x3000 + GSU_SCMR] & 0x18) == 0x18)
 	{
-	    FxEmulate ((Memory.FillRAM [0x3000 + GSU_CLSR] & 1) ? 1330 : 650);
+	    if (!Settings.WinterGold)
+		FxEmulate (~0);
+	    else
+		FxEmulate ((Memory.FillRAM [0x3000 + GSU_CLSR] & 1) ? 700 : 350);
 	    int GSUStatus = Memory.FillRAM [0x3000 + GSU_SFR] |
 			    (Memory.FillRAM [0x3000 + GSU_SFR + 1] << 8);
 	    if ((GSUStatus & (FLG_G | FLG_IRQ)) == FLG_IRQ)
 	    {
-		// XXX: Trigger a GSU IRQ.
+		// Trigger a GSU IRQ.
 		CPU.IRQActive |= GSU_IRQ_SOURCE;
 		CPU.Flags |= IRQ_PENDING_FLAG;
 	    }
@@ -2262,7 +2354,14 @@ void S9xSuperFXExec ()
 #else
     uint32 tmp =  (Memory.FillRAM[0x3034] << 16) + *(uint16 *) &Memory.FillRAM [0x301e];
 
-    if (tmp == -1) //0x01bf9e) //0x01bfc3) //0x09edaf) //-1) //0x57edaf)
+#if 0
+    if (tmp == 0x018428)
+    {
+	*(uint16 *) &SRAM [0x0064] = 0xbc00;
+	*(uint16 *) &SRAM [0x002c] = 0x8000;
+    }
+#endif
+    if (tmp == -1)//0x018428) //0x01bfc3) //0x09edaf) //-1) //0x57edaf)
     {
 	while (Memory.FillRAM [0x3030] & 0x20)
 	{
@@ -2374,7 +2473,12 @@ void S9xSuperFXExec ()
     }
     else
     {
-	FxEmulate (2000000);
+	uint32 t = (Memory.FillRAM [0x3034] << 16) +
+		   (Memory.FillRAM [0x301f] << 8) +
+		   (Memory.FillRAM [0x301e] << 0);
+
+printf ("%06x: %d\n", t, FxEmulate (2000000));
+//	FxEmulate (2000000);
     }
 #if 0
     if (!(CPU.Flags & TRACE_FLAG))
