@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -56,15 +56,16 @@
 #include "display.h"
 #include "apu.h"
 #include "soundux.h"
+#include "sa1.h"
 
 extern uint8 *SRAM;
 
 #ifdef ZSNES_FX
-extern "C" {
+START_EXTERN_C
 void S9xSuperFXPreSaveState ();
 void S9xSuperFXPostSaveState ();
 void S9xSuperFXPostLoadState ();
-}
+END_EXTERN_C
 #endif
 
 typedef struct {
@@ -322,7 +323,7 @@ static FreezeData SnapSoundData [] = {
     {OFFSET (channels [N].type), 4, INT_V}, \
     {OFFSET (channels [N].volume_left), 2, INT_V}, \
     {OFFSET (channels [N].volume_right), 2, INT_V}, \
-    {OFFSET (channels [N].frequency), 4, INT_V}, \
+    {OFFSET (channels [N].hertz), 4, INT_V}, \
     {OFFSET (channels [N].count), 4, INT_V}, \
     {OFFSET (channels [N].loop), 1, INT_V}, \
     {OFFSET (channels [N].envx), 4, INT_V}, \
@@ -351,6 +352,36 @@ static FreezeData SnapSoundData [] = {
 #undef O
 };
 
+#undef OFFSET
+#define OFFSET(f) Offset(f,struct SSA1Registers *)
+
+static FreezeData SnapSA1Registers [] = {
+    {OFFSET (PB),  1, INT_V},
+    {OFFSET (DB),  1, INT_V},
+    {OFFSET (P.W), 2, INT_V},
+    {OFFSET (A.W), 2, INT_V},
+    {OFFSET (D.W), 2, INT_V},
+    {OFFSET (S.W), 2, INT_V},
+    {OFFSET (X.W), 2, INT_V},
+    {OFFSET (Y.W), 2, INT_V},
+    {OFFSET (PC),  2, INT_V}
+};
+
+#undef OFFSET
+#define OFFSET(f) Offset(f,struct SSA1 *)
+
+static FreezeData SnapSA1 [] = {
+    {OFFSET (Flags), 4, INT_V},
+    {OFFSET (NMIActive), 1, INT_V},
+    {OFFSET (IRQActive), 1, INT_V},
+    {OFFSET (WaitingForInterrupt), 1, INT_V},
+    {OFFSET (op1), 2, INT_V},
+    {OFFSET (op2), 2, INT_V},
+    {OFFSET (arithmetic_op), 4, INT_V},
+    {OFFSET (sum), 8, INT_V},
+    {OFFSET (overflow), 1, INT_V}
+};
+
 static char ROMFilename [_MAX_PATH];
 static char SnapshotFilename [_MAX_PATH];
 
@@ -460,6 +491,14 @@ static void Freeze (STREAM stream)
 	FreezeStruct (stream, "SOU", &SoundData, SnapSoundData,
 		      COUNT (SnapSoundData));
     }
+    if (Settings.SA1)
+    {
+	SA1Registers.PC = SA1.PC - SA1.PCBase;
+	S9xSA1PackStatus ();
+	FreezeStruct (stream, "SA1", &SA1, SnapSA1, COUNT (SnapSA1));
+	FreezeStruct (stream, "SAR", &SA1Registers, SnapSA1Registers, 
+		      COUNT (SnapSA1Registers));
+    }
     S9xSetSoundMute (FALSE);
 #ifdef ZSNES_FX
     if (Settings.SuperFX)
@@ -551,6 +590,15 @@ static int Unfreeze (STREAM stream)
 	IAPU.APUExecuting = FALSE;
 	S9xSetSoundMute (TRUE);
     }
+    if ((result = UnfreezeStruct (stream, "SA1", &SA1, SnapSA1, 
+				  COUNT(SnapSA1))) == SUCCESS)
+    {
+	if ((result = UnfreezeStruct (stream, "SAR", &SA1Registers, 
+				      SnapSA1Registers, COUNT (SnapSA1Registers))) != SUCCESS)
+	    return (result);
+
+	S9xFixSA1AfterSnapshotLoad ();
+    }
     S9xFixSoundAfterSnapshotLoad ();
     ICPU.ShiftedPB = Registers.PB << 16;
     ICPU.ShiftedDB = Registers.DB << 16;
@@ -599,6 +647,7 @@ void FreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
     uint8 *ptr = block;
     uint16 word;
     uint32 dword;
+    int64  qword;
 
     // Build the block ready to be streamed out
     for (i = 0; i < num_fields; i++)
@@ -623,6 +672,17 @@ void FreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
 		*ptr++ = (uint8) (dword >> 8);
 		*ptr++ = (uint8) dword;
 		break;
+	    case 8:
+		qword = *((int64 *) ((uint8 *) base + fields [i].offset));
+		*ptr++ = (uint8) (qword >> 56);
+		*ptr++ = (uint8) (qword >> 48);
+		*ptr++ = (uint8) (qword >> 40);
+		*ptr++ = (uint8) (qword >> 32);
+		*ptr++ = (uint8) (qword >> 24);
+		*ptr++ = (uint8) (qword >> 16);
+		*ptr++ = (uint8) (qword >> 8);
+		*ptr++ = (uint8) qword;
+		break;
 	    }
 	    break;
 	case uint8_ARRAY_V:
@@ -683,10 +743,14 @@ int UnfreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
     uint8 *ptr = block;
     uint16 word;
     uint32 dword;
+    int64  qword;
     int result;
 
     if ((result = UnfreezeBlock (stream, name, block, len)) != SUCCESS)
+    {
+	delete block;
 	return (result);
+    }
 
     // Unpack the block of data into a C structure
     for (i = 0; i < num_fields; i++)
@@ -711,6 +775,17 @@ int UnfreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
 		dword |= *ptr++;
 		*((uint32 *) ((uint8 *) base + fields [i].offset)) = dword;
 		break;
+	    case 8:
+		qword  = (int64) *ptr++ << 56;
+		qword |= (int64) *ptr++ << 48;
+		qword |= (int64) *ptr++ << 40;
+		qword |= (int64) *ptr++ << 32;
+		qword |= (int64) *ptr++ << 24;
+		qword |= (int64) *ptr++ << 16;
+		qword |= (int64) *ptr++ << 8;
+		qword |= (int64) *ptr++;
+		*((int64 *) ((uint8 *) base + fields [i].offset)) = qword;
+		break;
 	    }
 	    break;
 	case uint8_ARRAY_V:
@@ -770,3 +845,40 @@ int UnfreezeBlock (STREAM stream, char *name, uint8 *block, int size)
 	
     return (SUCCESS);
 }
+
+bool8 S9xSPCDump (const char *filename)
+{
+    FILE *fs;
+
+    bool8 read_only = FALSE;
+    char def [PATH_MAX];
+    char title [PATH_MAX];
+    char drive [_MAX_DRIVE];
+    char dir [_MAX_DIR];
+    char ext [_MAX_EXT];
+
+    S9xSetSoundMute (TRUE);
+
+    if (!(fs = fopen (filename, "wb")))
+	return (FALSE);
+
+    if (fputs ("SNES-SPC700 Sound File Data 0.10", fs) == EOF ||
+	fseek (fs, 37, SEEK_SET) == EOF ||
+	fwrite (&APURegisters.PC, 2, 1, fs) == EOF ||
+	fputc (APURegisters.YA.B.A, fs) == EOF ||
+	fputc (APURegisters.X, fs) == EOF ||
+	fputc (APURegisters.YA.B.Y, fs) == EOF ||
+	fputc (APURegisters.P, fs) == EOF ||
+	fputc (APURegisters.S, fs) == EOF ||
+	fseek (fs, 256, SEEK_SET) == EOF ||
+	fwrite (IAPU.RAM, 1, 0x10000, fs) == EOF ||
+	fwrite (APU.DSP, 1, 192, fs) == EOF ||
+	fwrite (APU.ExtraRAM, 1, 64, fs) == EOF ||
+	fclose (fs) < 0)
+    {
+	S9xSetSoundMute (FALSE);
+	return (FALSE);
+    }
+    S9xSetSoundMute (FALSE);
+    return (TRUE);
+}
