@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -166,6 +166,15 @@ compressed with gzip or compress.\n");
     exit (1);
 }
 
+#ifdef STORM
+extern int dofps;
+extern int hicolor;
+extern int secondjoy;
+extern int minimal;
+int prelude=0;
+extern int unit;
+#endif
+
 char *S9xParseArgs (char **argv, int argc)
 {
     char *rom_filename = NULL;
@@ -459,6 +468,32 @@ char *S9xParseArgs (char **argv, int argc)
 	    {
 		Settings.DisableMasterVolume = TRUE;
 	    }
+	    else if (strcasecmp (argv [i], "-soundsync") == 0 ||
+		     strcasecmp (argv [i], "-sy") == 0)
+	    {
+		Settings.SoundSync = TRUE;
+		Settings.SoundEnvelopeHeightReading = TRUE;
+		Settings.InterpolatedSound = TRUE;
+	    }
+	    else if (strcasecmp (argv [i], "-soundsync2") == 0 ||
+		     strcasecmp (argv [i], "-sy2") == 0)
+	    {
+		Settings.SoundSync = 2;
+		Settings.SoundEnvelopeHeightReading = TRUE;
+		Settings.InterpolatedSound = TRUE;
+	    }
+	    else if (strcasecmp (argv [i], "-interpolatedsound") == 0 ||
+		     strcasecmp (argv [i], "-is") == 0)
+	    {
+		Settings.InterpolatedSound = TRUE;
+	    }
+#ifdef USE_THREADS
+	    else if (strcasecmp (argv [i], "-threadsound") == 0 ||
+		     strcasecmp (argv [i], "-ts") == 0)
+	    {
+		Settings.ThreadSound = TRUE;
+	    }
+#endif
 	    else if (strcasecmp (argv [i], "-nosuperfx") == 0 ||
 		     strcasecmp (argv [i], "-nosfx") == 0)
 		Settings.ForceNoSuperFX = TRUE;
@@ -529,6 +564,21 @@ char *S9xParseArgs (char **argv, int argc)
 		Settings.NetPlay = TRUE;
 	    }
 #endif
+#ifdef STORM
+            else if (strcasecmp(argv[i],"-nosecondjoy")==0){secondjoy=0;}
+            else if (strcasecmp(argv[i],"-showfps")==0){dofps=1;}
+            else if (strcasecmp(argv[i],"-hicolor")==0){hicolor=1;}
+            else if (strcasecmp(argv[i],"-minimal")==0){minimal=1;printf("Keyboard with exception of ESC switched off!\n");}
+            else if (strcasecmp(argv[i],"-ahiunit")==0)
+            {
+             if (i+1<argc)
+             {
+              fprintf(stderr,"AHI Unit set to: Unit %i\n",atoi(argv[++i]));
+              unit=atoi(argv[++i]);
+             }
+            }
+#endif
+
 	    else
 		S9xParseArg (argv, i, argc);
 	}
