@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -42,7 +42,7 @@
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-//#include <sys/io.h>
+#include <sys/io.h>
 #include <stdlib.h>
 #include <signal.h>
 
@@ -265,8 +265,11 @@ void S9xInitDisplay (int argc, char **argv)
 	GFX.Screen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	GFX.SubScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	DeltaScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.ZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
+	GFX.SubZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
 
-	if (!GFX.Screen || !GFX.SubScreen || !DeltaScreen)
+	if (!GFX.Screen || !GFX.SubScreen || !DeltaScreen || 
+	    !GFX.ZBuffer || !GFX.SubZBuffer)
 	{
 	    fprintf (stdout, "Cannot allocate screen buffer.\n");
 	    S9xExit ();
@@ -278,6 +281,8 @@ void S9xInitDisplay (int argc, char **argv)
 	GFX.Pitch = IMAGE_WIDTH * 2;
 	GFX.Screen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	GFX.SubScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.ZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
+	GFX.SubZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
 
 	if (!GFX.Screen || !GFX.SubScreen)
 	{
@@ -301,6 +306,8 @@ void S9xInitDisplay (int argc, char **argv)
 	    fprintf (stdout, "Cannot allocate shadow screen buffer.\n");
 	    S9xExit ();
         }
+	GFX.ZBuffer = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.SubZBuffer = NULL;
     }
     ZeroMemory (GFX.Screen, GFX.Pitch * IMAGE_HEIGHT);
     if (GFX.SubScreen)
@@ -327,10 +334,15 @@ void S9xDeinitDisplay ()
 	free ((char *) GFX.SubScreen);
     if (DeltaScreen)
 	free ((char *) DeltaScreen);
-
+    if (GFX.ZBuffer)
+	free ((char *) GFX.ZBuffer);
+    if (GFX.SubZBuffer)
+	free ((char *) GFX.SubZBuffer);
     GFX.Screen = NULL;
     GFX.SubScreen = NULL;
     DeltaScreen = NULL;
+    GFX.ZBuffer = NULL;
+    GFX.SubZBuffer = NULL;
 }
 
 void S9xSetPalette ()
@@ -666,7 +678,7 @@ void S9xPutImage (int width, int height)
 		for (int y = y_start; y < y_end; y++, s += GFX.Pitch, p += screen_pitch)
 		    memcpy (p, s, width);
 #else
-		gl_putbox (0, 0, width, height, GFX.Screen);
+		gl_putbox (0, 0, width * 2, height * 2, GFX.Screen);
 #endif
 	    }
 	}
