@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -48,16 +48,27 @@
 #include <sys/types.h>
 #include <ctype.h>
 
+
 #if defined(__linux) || defined(__sun)
+#undef USE_THREADS
+#define USE_THREADS
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
-#include <signal.h>
+#endif
+
+#ifdef USE_THREADS
+#include <pthread.h>
+#include <sched.h>
+
+pthread_t thread;
+pthread_mutex_t mutex;
 #endif
 
 #ifdef __linux
 #include <sys/soundcard.h>
+#include <sys/mman.h>
 #endif
 
 #ifdef __sun
@@ -88,7 +99,7 @@ typedef void (*SIG_PF)();
 #endif
 
 uint32 joypads [5] = {0};
-int NumControllers = 2;
+int NumControllers = 5;
 
 #ifdef JOYSTICK_SUPPORT
 #ifdef __linux
@@ -143,6 +154,7 @@ void ReadJoysticks ();
 #endif
 
 void InitTimer ();
+void *S9xProcessSound (void *);
 
 char *rom_filename = NULL;
 char *snapshot_filename = NULL;
@@ -288,7 +300,6 @@ int main (int argc, char **argv)
     if (argc < S9xMinCommandLineArgs ())
 	S9xUsage ();
 	
-//printf ("/dev/dsp: %d\n", open ("/dev/dsp", O_WRONLY));
     ZeroMemory (&Settings, sizeof (Settings));
 
 #ifdef JOYSTICK_SUPPORT
@@ -320,6 +331,7 @@ int main (int argc, char **argv)
     Settings.SupportHiRes = FALSE;
     Settings.NetPlay = FALSE;
     Settings.ServerName [0] = 0;
+    Settings.ThreadSound = FALSE;
 #ifdef NETPLAY_SUPPORT
     Settings.Port = NETPLAY_DEFAULT_PORT;
 #endif
@@ -551,6 +563,7 @@ Connected to \"%s\" on port %d as player #%d playing \"%s\"\n",
 
 void S9xExit ()
 {
+    S9xSetSoundMute (TRUE);
     S9xDeinitDisplay ();
     Memory.SaveSRAM (S9xGetSRAMFilename ());
     Memory.Deinit ();
@@ -574,6 +587,7 @@ void S9xInitInputDevices ()
 #ifdef JOYSTICK_SUPPORT
 void InitJoysticks ()
 {
+#ifdef JSIOCGVERSION
     int version;
     unsigned char axes, buttons;
 
@@ -582,6 +596,7 @@ void InitJoysticks ()
 	perror (js_device [0]);
 	return;
     }
+
     if (ioctl (js_fd [0], JSIOCGVERSION, &version))
     {
         puts("joystick: You need at least driver version 1.0 for joystick support");
@@ -614,15 +629,17 @@ void InitJoysticks ()
 	{
 	    ioctl (js_fd [0], JSIOCGAXES, &axes);
 	    ioctl (js_fd [0], JSIOCGBUTTONS, &buttons);
-	    printf ("and %d-axis %d-button (%s) as pad2", axes, buttons, js_device [1]);
+	    printf (" and %d-axis %d-button (%s) as pad2", axes, buttons, js_device [1]);
 	}
     }
 
     puts (".");
+#endif
 }
 
 void ReadJoysticks ()
 {
+#ifdef JSIOCGVERSION
     struct js_event js_ev;
     int i;
 
@@ -686,6 +703,7 @@ void ReadJoysticks ()
 	    }
 	}
     }
+#endif
 }
 #endif // defined (JOYSTICK_SUPPORT)
 
@@ -850,16 +868,6 @@ bool8 S9xInitUpdate ()
 bool8 S9xDeinitUpdate (int Width, int Height, bool8 /*sixteen_bit*/)
 {
     S9xPutImage (Width, Height);
-#if 0
-    extern int plot2_count, plot4_count, plot8_count, plotobj_count;
-
-    printf ("\r%5d, %5d, %5d, %5d", plot2_count, plot4_count, plot8_count, plotobj_count);
-    plot2_count = 0;
-    plot4_count = 0;
-    plot8_count = 0;
-    plotobj_count = 0;
-    fflush (stdout);
-#endif
     return (TRUE);
 }
 
@@ -959,8 +967,8 @@ void S9xToggleSoundChannel (int c)
 
 static void SoundTrigger ()
 {
-    if (Settings.APUEnabled)
-	CPU.Flags |= PROCESS_SOUND_FLAG;
+    if (Settings.APUEnabled && !so.mute_sound)
+	S9xProcessSound (NULL);
 }
 
 void StopTimer ()
@@ -980,6 +988,15 @@ void InitTimer ()
     struct itimerval timeout;
     struct sigaction sa;
     
+#ifdef USE_THREADS
+    if (Settings.ThreadSound)
+    {
+	pthread_mutex_init (&mutex, NULL);
+	pthread_create (&thread, NULL, S9xProcessSound, NULL);
+	return;
+    }
+#endif
+
 #if defined(__linux)
     sa.sa_handler = (SIG_PF) SoundTrigger;
 #else
@@ -1037,6 +1054,12 @@ void S9xSyncSpeed ()
     }
     else
 #endif
+    if (Settings.SoundSync == 2)
+    {
+	IPPU.RenderThisFrame = TRUE;
+	IPPU.SkippedFrames = 0;
+    }
+    else
     if (Settings.SkipFrames == AUTO_FRAMERATE)
     {
 	static struct timeval next1 = {0};
@@ -1055,12 +1078,15 @@ void S9xSyncSpeed ()
 
 	if (timercmp(&next1, &now, >))
 	{
-	    do
+	    if (IPPU.SkippedFrames == 0)
 	    {
-		CHECK_SOUND ();
-		S9xProcessEvents (FALSE);
-		while (gettimeofday (&now, NULL) < 0) ;
-	    } while (timercmp(&next1, &now, >));
+		do
+		{
+		    CHECK_SOUND ();
+		    S9xProcessEvents (FALSE);
+		    while (gettimeofday (&now, NULL) < 0) ;
+		} while (timercmp(&next1, &now, >));
+	    }
 	    IPPU.RenderThisFrame = TRUE;
 	    IPPU.SkippedFrames = 0;
 	}
@@ -1392,7 +1418,28 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
 	return (FALSE);
     }
 
+#ifdef MMAP_SOUND 
+   if (ioctl (so.sound_fd, SNDCTL_DSP_GETCAPS, &J) < 0)
+    {
+	perror ("ioctl SNDCTL_DSP_GETCAPS");
+    }
+    else
+    {
+	if (J & DSP_CAP_MMAP)
+	    printf ("DSP_CAP_MMAP supported\n");
+    }
+    void *ptr;
+
+    if ((ptr = mmap (0, so.buffer_size * 3, PROT_WRITE, 0, so.sound_fd, 0)) == 0)
+	printf ("mmap failed\n");
+
+    J = 0;
+    if (ioctl (so.sound_fd, SNDCTL_DSP_SETTRIGGER, &J) < 0)
+	perror ("ioctl SNDCTL_DSP_SETTRIGGER");
+#endif
+
     J = AFMT_S16_LE;
+//    J = AFMT_U8;
     if (ioctl (so.sound_fd, SNDCTL_DSP_SETFMT, &J) < 0)
     {
 	perror ("ioctl SNDCTL_DSP_SETFMT");
@@ -1437,7 +1484,7 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
 	buffer_size *= 2;
 
     int power2 = log2 (buffer_size);
-    J = K = power2 | (SOUND_BUFS << 16);
+    J = K = power2 | (3 << 16);
     if (ioctl (so.sound_fd, SNDCTL_DSP_SETFRAGMENT, &J) < 0)
     {
 	perror ("ioctl SNDCTL_DSP_SETFRAGMENT");
@@ -1445,6 +1492,20 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
     }
     ioctl (so.sound_fd, SNDCTL_DSP_GETBLKSIZE, &so.buffer_size);
     
+#ifdef MMAP_SOUND
+    J = PCM_ENABLE_OUTPUT;
+    if (ioctl (so.sound_fd, SNDCTL_DSP_SETTRIGGER, &J) < 0)
+	perror ("ioctl SNDCTL_DSP_SETTRIGGER");
+#endif
+
+#if 0
+    buffmem_desc buff;
+    buff.size = so.buffer_size * 3;
+    buff.buffer = ptr;
+    if (ioctl (so.sound_fd, SNDCTL_DSP_MAPOUTBUF, &buff) < 0)
+	perror ("ioctl SNDCTL_DSP_MAPOUTBUF");
+#endif
+
     printf ("Rate: %d, Buffer size: %d, 16-bit: %s, Stereo: %s, Encoded: %s\n",
 	    so.playback_rate, so.buffer_size, so.sixteen_bit ? "yes" : "no",
 	    so.stereo ? "yes" : "no", so.encoded ? "yes" : "no");
@@ -1456,49 +1517,232 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
 #if defined (__linux) || defined (__sun)
 void S9xUnixProcessSound (void)
 {
-    uint8 Buf[MAX_BUFFER_SIZE];
+}
 
-    CPU.Flags &= ~PROCESS_SOUND_FLAG;
-    int sample_count = so.buffer_size;
-    if (so.sixteen_bit)
-	sample_count >>= 1;
+static uint8 Buf[MAX_BUFFER_SIZE];
 
+#define FIXED_POINT 0x10000
+#define FIXED_POINT_SHIFT 16
+#define FIXED_POINT_REMAINDER 0xffff
+
+static volatile bool8 block_signal = FALSE;
+static volatile bool8 block_generate_sound = FALSE;
+static volatile bool8 pending_signal = FALSE;
+
+void S9xGenerateSound ()
+{
+    int bytes_so_far = so.sixteen_bit ? (so.samples_mixed_so_far << 1) :
+				        so.samples_mixed_so_far;
+    if (Settings.SoundSync == 2)
+    {
+	// Assumes sound is signal driven
+	while (so.samples_mixed_so_far >= so.buffer_size && !so.mute_sound)
+	    pause ();
+    }
+    else
+    if (bytes_so_far >= so.buffer_size)
+	return;
+
+#ifdef USE_THREADS
+    if (Settings.ThreadSound)
+    {
+	if (block_generate_sound || pthread_mutex_trylock (&mutex))
+	    return;
+    }
+#endif
+
+    block_signal = TRUE;
+
+    so.err_counter += so.err_rate;
+    if (so.err_counter >= FIXED_POINT)
+    {
+        int sample_count = so.err_counter >> FIXED_POINT_SHIFT;
+	int byte_offset;
+	int byte_count;
+
+        so.err_counter &= FIXED_POINT_REMAINDER;
+	if (so.stereo)
+	    sample_count <<= 1;
+	byte_offset = bytes_so_far + so.play_position;
+	    
+	do
+	{
+	    int sc = sample_count;
+	    byte_count = sample_count;
+	    if (so.sixteen_bit)
+		byte_count <<= 1;
+	    
+	    if ((byte_offset & SOUND_BUFFER_SIZE_MASK) + byte_count > SOUND_BUFFER_SIZE)
+	    {
+		sc = SOUND_BUFFER_SIZE - (byte_offset & SOUND_BUFFER_SIZE_MASK);
+		byte_count = sc;
+		if (so.sixteen_bit)
+		    sc >>= 1;
+	    }
+	    if (bytes_so_far + byte_count > so.buffer_size)
+	    {
+		byte_count = so.buffer_size - bytes_so_far;
+		if (byte_count == 0)
+		    break;
+		sc = byte_count;
+		if (so.sixteen_bit)
+		    sc >>= 1;
+	    }
+	    S9xMixSamplesO (Buf, sc,
+			    byte_offset & SOUND_BUFFER_SIZE_MASK);
+	    so.samples_mixed_so_far += sc;
+	    sample_count -= sc;
+	    bytes_so_far = so.sixteen_bit ? (so.samples_mixed_so_far << 1) :
+	 	           so.samples_mixed_so_far;
+	    byte_offset += byte_count;
+	} while (sample_count > 0);
+    }
+    block_signal = FALSE;
+
+#ifdef USE_THREADS
+    if (Settings.ThreadSound)
+	pthread_mutex_unlock (&mutex);
+    else
+#endif    
+    if (pending_signal)
+    {
+	S9xProcessSound (NULL);
+	pending_signal = FALSE;
+    }
+}
+
+void *S9xProcessSound (void *)
+{
 #ifdef __linux
     audio_buf_info info;
-    if (ioctl (so.sound_fd, SNDCTL_DSP_GETOSPACE, &info) == -1 ||
-	info.bytes < so.buffer_size)
-	return;
+    count_info count;
+
+    if (!Settings.ThreadSound &&
+	(ioctl (so.sound_fd, SNDCTL_DSP_GETOSPACE, &info) == -1 ||
+	 info.bytes < so.buffer_size))
+	return (NULL);
+
+#ifdef MMAP_SOUND
+    if (ioctl (so.sound_fd, SNDCTL_DSP_GETOPTR, &count) < 0)
+    {
+	printf ("F"); fflush (stdout);
+    }
+    else
+    {
+	printf ("<%d,%d>", count.blocks, count.bytes); fflush (stdout);
+    	return (NULL);
+    }
+#endif
+
 #endif
 #ifdef __sun
     audio_info_t audio;
-    if (ioctl (so.sound_fd, AUDIO_GETINFO, &audio) < 0)
-	return;
+    if (!Settings.ThreadSound)
+    {
+	if (ioctl (so.sound_fd, AUDIO_GETINFO, &audio) < 0)
+	    return (NULL);
 
-    if (audio.play.eof == 0)
-	so.last_eof = -2;
-    else
-    if (audio.play.eof == so.last_eof)
-	return;
+	if (audio.play.eof == 0)
+	    so.last_eof = -2;
+	else
+	if (audio.play.eof == so.last_eof)
+	    return (NULL);
 
-    so.last_eof++;
+	so.last_eof++;
+    }
 #endif
 
-    S9xMixSamples (Buf, sample_count);
+#ifdef USE_THREADS
+    do
+    {
+#endif
 
-    if (!so.mute_sound)
+    int sample_count = so.buffer_size;
+    int byte_offset;
+
+    if (so.sixteen_bit)
+	sample_count >>= 1;
+ 
+#ifdef USE_THREADS
+    if (Settings.ThreadSound)
+	pthread_mutex_lock (&mutex);
+    else
+#endif
+    if (block_signal)
+    {
+	pending_signal = TRUE;
+	return (NULL);
+    }
+
+    block_generate_sound = TRUE;
+
+    if (so.samples_mixed_so_far < sample_count)
+    {
+	byte_offset = so.play_position + 
+		      (so.sixteen_bit ? (so.samples_mixed_so_far << 1)
+				      : so.samples_mixed_so_far);
+
+//printf ("%d:", sample_count - so.samples_mixed_so_far); fflush (stdout);
+	if (Settings.SoundSync == 2)
+	{
+	    /*memset (Buf + (byte_offset & SOUND_BUFFER_SIZE_MASK), 0,
+		    sample_count - so.samples_mixed_so_far);*/
+	}
+	else
+	    S9xMixSamplesO (Buf, sample_count - so.samples_mixed_so_far,
+			    byte_offset & SOUND_BUFFER_SIZE_MASK);
+	so.samples_mixed_so_far = 0;
+    }
+    else
+	so.samples_mixed_so_far -= sample_count;
+    
+//    if (!so.mute_sound)
     {
-	int J = 0;
 	int I;
+	int J = so.buffer_size;
+
+	byte_offset = so.play_position;
+	so.play_position = (so.play_position + so.buffer_size) & SOUND_BUFFER_SIZE_MASK;
+
+#ifdef USE_THREADS
+	if (Settings.ThreadSound)
+	    pthread_mutex_unlock (&mutex);
+#endif
+	block_generate_sound = FALSE;
 	do
 	{
-	    I = write (so.sound_fd, (char *) Buf + J, so.buffer_size - J);
-	    if (I > 0)
-		J += I;
-	} while ((I < 0 && errno == EINTR) || J < so.buffer_size);
+	    if (byte_offset + J > SOUND_BUFFER_SIZE)
+	    {
+		I = write (so.sound_fd, (char *) Buf + byte_offset,
+			   SOUND_BUFFER_SIZE - byte_offset);
+		if (I > 0)
+		{
+		    J -= I;
+		    byte_offset = (byte_offset + I) & SOUND_BUFFER_SIZE_MASK;
+		}
+	    }
+	    else
+	    {
+		I = write (so.sound_fd, (char *) Buf + byte_offset, J);
+		if (I > 0)
+		{
+		    J -= I;
+		    byte_offset = (byte_offset + I) & SOUND_BUFFER_SIZE_MASK;
+		}
+	    }
+	} while ((I < 0 && errno == EINTR) || J > 0);
     }
-#ifdef __sun
-    write (so.sound_fd, NULL, 0);
+
+#ifdef USE_THREADS
+    } while (Settings.ThreadSound);
 #endif
+
+#ifdef __sun
+    if (!Settings.ThreadSound)
+	write (so.sound_fd, NULL, 0);
+#endif
+
+    return (NULL);
 }
 #endif
 
