@@ -1,12 +1,12 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
+ *                                      Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998 Ivar (Ivar@snes9x.com) and
- *                          Gary Henderson.
+ * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
+ *                                Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
@@ -42,6 +42,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <errno.h>
 
 #include "snes9x.h"
 #include "memmap.h"
@@ -51,19 +52,32 @@
 #include "gfx.h"
 #include "display.h"
 #include "apu.h"
+#include "soundux.h"
+
+#if 0
+#define QT_CLEAN_NAMESPACE
+#include <qapplication.h>
+#include "snes9x_gui.h"
+#endif
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
 
+#ifdef USE_GLIDE
+void S9xGlideEnable (bool8);
+void S9xGlidePutImage (int, int);
+void S9xSwitchToGlideMode (bool8);
+#endif
+
 #ifdef USE_DGA_EXTENSION
 #include <X11/extensions/xf86dga.h>
 
 void CreateFullScreenWindow ();
 void S9xSwitchToFullScreen (bool8 enable);
 static Window fs_window = 0;
-static scale = TRUE;
+static bool8 scale = TRUE;
 
 #ifdef USE_VIDMODE_EXTENSION
 #if defined (__cplusplus) || defined (c_plusplus)
@@ -96,6 +110,7 @@ typedef struct
     XF86VidModeModeInfo orig;
     XF86VidModeModeInfo *best;
     bool8 no_mode_switch;
+    bool8 start_full_screen;
 #endif
 } XF86Data;
 
@@ -149,9 +164,15 @@ static int mouse_offset_y = 0;
 
 static XColor fixed_colours [256] = { {0} };
 static uint8 palette [0x10000] = {0};
+static char InfoString [256];
 
 extern uint32 joypads [5];
 
+#if 0
+QApplication *app;
+Snes9xGUI *gui;
+#endif
+
 #ifdef MITSHM
 #include <sys/ipc.h>
 #include <sys/shm.h>
@@ -171,9 +192,14 @@ void Convert16To24Packed (int width, int height);
 void SetupImage ();
 int ErrorHandler (Display *, XErrorEvent *);
 void TVMode (int width, int height);
+void S9xDisplayStateChange (const char *str, bool8 on);
 
 void S9xDeinitDisplay ()
 {
+#ifdef USE_GLIDE
+    S9xGlideEnable (FALSE);
+#endif
+
     S9xTextMode ();
     if (InterpolateScreen &&
 	(!image || InterpolateScreen != (uint8 *) image->data))
@@ -191,6 +217,14 @@ void S9xDeinitDisplay ()
 
     GFX.SubScreen = NULL;
 
+    if (GFX.ZBuffer)
+	free ((char *) GFX.ZBuffer);
+    GFX.ZBuffer = NULL;
+
+    if (GFX.SubZBuffer)
+	free ((char *) GFX.SubZBuffer);
+    GFX.SubZBuffer = NULL;
+
     if (image)
     {
 #ifdef MITSHM
@@ -260,7 +294,13 @@ Snes9x needs an X Window System server set to 8, 15, 16, 24 or 32-bit colour dep
 supporting PseudoColor, TrueColor or GrayScale.\n");
 	exit (1);
     }
+
+#ifdef USE_GLIDE
+    // Allways enable transparency effects and 16-bit rendering when using the
+    // 3D/FX chip as the output image blitter.
+#else
     if (depth >= 15 && !Settings.ForceNoTransparency)
+#endif
     {
 	Settings.Transparency = TRUE;
 	Settings.SixteenBit = TRUE;
@@ -502,6 +542,16 @@ supporting PseudoColor, TrueColor or GrayScale.\n");
 	bytes_per_pixel = 4;
 	break;
     }
+#if 0
+    app = new QApplication (display);
+    gui = new Snes9xGUI ();
+    app->setMainWidget (gui);
+    gui->show ();
+#endif
+#ifdef USE_GLIDE
+    putenv("FX_GLIDE_NO_SPLASH=");
+    S9xSwitchToGlideMode (TRUE);
+#endif
 }
 
 void SetupImage ()
@@ -546,6 +596,14 @@ void SetupImage ()
 
     GFX.SubScreen = NULL;
 
+    if (GFX.ZBuffer)
+	free ((char *) GFX.ZBuffer);
+    GFX.ZBuffer = NULL;
+
+    if (GFX.SubZBuffer)
+	free ((char *) GFX.SubZBuffer);
+    GFX.SubZBuffer = NULL;
+
     if (image)
     {
 #ifdef MITSHM
@@ -664,13 +722,17 @@ void SetupImage ()
 	    GFX.Pitch = image->bytes_per_line;
 	}
 	GFX.SubScreen = NULL;
+	GFX.ZBuffer = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.SubZBuffer = NULL;
     }
     else
     if (depth == 8)
     {
-	GFX.Screen = (uint8 *) malloc ((IMAGE_WIDTH + 2) * IMAGE_HEIGHT * 2);
 	GFX.Pitch = (IMAGE_WIDTH + 2) * 2;
-	GFX.SubScreen = (uint8 *) malloc ((IMAGE_WIDTH + 2) * IMAGE_HEIGHT * 2);
+	GFX.Screen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.SubScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.ZBuffer = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.SubZBuffer = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	if (interpolate)
 	{
 	    InterpolateScreen = (uint8 *) malloc (514 * 478 * 2);
@@ -681,8 +743,8 @@ void SetupImage ()
     {
 	if ((depth != 15 && depth != 16) || interpolate)
 	{
-	    GFX.Screen = (uint8 *) malloc ((IMAGE_WIDTH + 2) * IMAGE_HEIGHT * 2);
 	    GFX.Pitch = (IMAGE_WIDTH + 2) * 2;
+	    GFX.Screen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	    if (interpolate)
 	    {
 		if (ImageNeedsScaling || (depth != 15 && depth != 16)
@@ -690,8 +752,8 @@ void SetupImage ()
 		    || xf86data.scale
 #endif
 )
-		    InterpolateScreen = (uint8 *) malloc (514 * 478 * 2);
-		DeltaScreen = (uint8 *) malloc ((IMAGE_WIDTH + 2) * IMAGE_HEIGHT * 2);
+		InterpolateScreen = (uint8 *) malloc (514 * 478 * 2);
+		DeltaScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
 	    }
 	}
 	else
@@ -700,6 +762,8 @@ void SetupImage ()
 	    GFX.Pitch = image->bytes_per_line;
 	}
 	GFX.SubScreen = (uint8 *) malloc (GFX.Pitch * IMAGE_HEIGHT);
+	GFX.ZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
+	GFX.SubZBuffer = (uint8 *) malloc ((GFX.Pitch >> 1) * IMAGE_HEIGHT);
     }
     GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;
     ZeroMemory (GFX.Screen, GFX.Pitch * IMAGE_HEIGHT);
@@ -715,7 +779,9 @@ void SetupImage ()
 
 int ErrorHandler (Display *, XErrorEvent *)
 {
+#ifdef MITSHM
     UseSHM = 0;
+#endif
     return (0);
 }
 
@@ -754,6 +820,13 @@ void S9xProcessEvents (bool8 block)
 	XNextEvent (display, &event);
 	block = FALSE;
 
+#if 0
+	if (event.xany.window != window)
+	{
+	    app->x11ProcessEvent (&event);
+	    continue;
+	}
+#endif
 	uint8 byte1 = 0;
 	uint8 byte2 = 0;
 	uint8 byte3 = 0;
@@ -834,46 +907,82 @@ void S9xProcessEvents (bool8 block)
 
 	    case XK_0:
 		if (event.type == KeyPress)
+		{
 		    Settings.DisableHDMA = !Settings.DisableHDMA;
+		    S9xDisplayStateChange ("HDMA emulation", !Settings.DisableHDMA);
+		}
 		break;
 	    case XK_1:
 		if (event.type == KeyPress)
+		{
 		    PPU.BG_Forced ^= 1;
+		    S9xDisplayStateChange ("BG#0", !(PPU.BG_Forced & 1));
+		}
 		break;
 	    case XK_2:
 		if (event.type == KeyPress)
+		{
 		    PPU.BG_Forced ^= 2;
+		    S9xDisplayStateChange ("BG#1", !(PPU.BG_Forced & 2));
+		}
 		break;
 	    case XK_3:
 		if (event.type == KeyPress)
+		{
 		    PPU.BG_Forced ^= 4;
+		    S9xDisplayStateChange ("BG#2", !(PPU.BG_Forced & 4));
+		}
 		break;
 	    case XK_4:
 		if (event.type == KeyPress)
+		{
 		    PPU.BG_Forced ^= 8;
+		    S9xDisplayStateChange ("BG#3", !(PPU.BG_Forced & 8));
+		}
 		break;
 	    case XK_5:
 		if (event.type == KeyPress)
+		{
 		    PPU.BG_Forced ^= 16;
+		    S9xDisplayStateChange ("Sprites", !(PPU.BG_Forced & 16));
+		}
 		break;
 	    case XK_6:
 		if (event.type == KeyPress)
+		{
 		    Settings.SwapJoypads = !Settings.SwapJoypads;
+		    S9xDisplayStateChange ("Joypad swapping", Settings.SwapJoypads);
+		}
+		break;
+	    case XK_7:
+		if (event.type == KeyPress)
+		{
+		    static char *controllers [] = {
+			"Multiplayer 5 on #0", "Joypad on #0", "Mouse on #1",
+			"Mouse on #0", "Superscope on #1"
+		    };
+		    S9xNextController ();
+		    S9xSetInfoString (controllers [IPPU.Controller]);
+		}
+		break;
+
+	    case XK_8:
+		if (event.type == KeyPress)
+		{
+		    Settings.BGLayering = !Settings.BGLayering;
+		    S9xDisplayStateChange ("Background layering hack", 
+					   Settings.BGLayering);
+		}
 		break;
 	    case XK_9:
 		if (event.type == KeyPress)
 		    if (Settings.SixteenBit)
+		    {
 			Settings.Transparency = !Settings.Transparency;
+			S9xDisplayStateChange ("Transparency effects", 
+					       Settings.Transparency);
+		    }
 		break;
-	    case XK_8:
-		if (event.type == KeyPress)
-		    Settings.BGLayering = !Settings.BGLayering;
-		break;
-	    case XK_7:
-		if (event.type == KeyPress)
-		    S9xNextController ();
-		break;
-
 	    case XK_minus:
 		if (event.type == KeyPress)
 		{
@@ -882,14 +991,27 @@ void S9xProcessEvents (bool8 block)
 			// Decrease emulated frame time by 1ms
 			if (Settings.FrameTime >= 1000)
 			    Settings.FrameTime -= 1000;
+
+			sprintf (InfoString, "Emulated frame time: %dms",
+				 Settings.FrameTime / 1000);
+			S9xSetInfoString (InfoString);
 		    }
 		    else
 		    {
 			if (Settings.SkipFrames <= 1)
 			    Settings.SkipFrames = AUTO_FRAMERATE;
 			else
-			if (Settings.SkipFrames != AUTO_FRAMERATE)
-			    Settings.SkipFrames--;
+			    if (Settings.SkipFrames != AUTO_FRAMERATE)
+				Settings.SkipFrames--;
+
+			if (Settings.SkipFrames == AUTO_FRAMERATE)
+			    S9xSetInfoString ("Auto frame skip");
+			else
+			{
+			    sprintf (InfoString, "Frame skip: %d",
+				     Settings.SkipFrames - 1);
+			    S9xSetInfoString (InfoString);
+			}
 		    }
 		}
 		break;
@@ -902,6 +1024,9 @@ void S9xProcessEvents (bool8 block)
 		    {
 			// Increase emulated frame time by 1ms
 			Settings.FrameTime += 1000;
+			sprintf (InfoString, "Emulated frame time: %dms",
+				 Settings.FrameTime / 1000);
+			S9xSetInfoString (InfoString);
 		    }
 		    else
 		    {
@@ -910,19 +1035,51 @@ void S9xProcessEvents (bool8 block)
 			else
 			if (Settings.SkipFrames < 10)
 			    Settings.SkipFrames++;
+
+			if (Settings.SkipFrames == AUTO_FRAMERATE)
+			    S9xSetInfoString ("Auto frame skip");
+			else
+			{
+			    sprintf (InfoString, "Frame skip: %d",
+				     Settings.SkipFrames - 1);
+			    S9xSetInfoString (InfoString);
+			}
 		    }
 		}
 		break;
 
 	    case XK_BackSpace:
 		if (event.type == KeyPress)
+		{
 		    Settings.DisableGraphicWindows = !Settings.DisableGraphicWindows;
+		    S9xDisplayStateChange ("Graphic clip windows",
+					   !Settings.DisableGraphicWindows);
+		}
+		break;
+	    case XK_bracketleft:
+		if (event.type == KeyPress)
+		{
+		    Settings.InterpolatedSound ^= 1;
+		    S9xDisplayStateChange ("Interpolated sound",
+					   Settings.InterpolatedSound);
+		}
+		break;
+	    case XK_bracketright:
+		if (event.type == KeyPress)
+		{
+		    Settings.SoundSync ^= 1;
+		    S9xDisplayStateChange ("Synchronised sound",
+					   Settings.SoundSync);
+		}
 		break;
 	    case XK_Scroll_Lock:
 	    case XK_Pause:
 	    case XK_Break:
 		if (event.type == KeyPress)
+		{
 		    Settings.Paused ^= 1;
+		    S9xDisplayStateChange ("Pause", Settings.Paused);
+		}
 		break;
 
 	    case XK_Tab:
@@ -944,20 +1101,34 @@ void S9xProcessEvents (bool8 block)
 		break;
 #endif
 	    case XK_F1:
-#ifdef DEBUGGER
 		if (event.type == KeyPress && 
 		    (event.xkey.state & (Mod1Mask | ControlMask)))
 		{
-		    CPU.Flags |= DEBUG_MODE_FLAG;
+		    char def [PATH_MAX];
+		    char filename [PATH_MAX];
+		    char drive [_MAX_DRIVE];
+		    char dir [_MAX_DIR];
+		    char ext [_MAX_EXT];
+
+		    _splitpath (Memory.ROMFilename, drive, dir, def, ext);
+		    strcpy (ext, "spc");
+		    _makepath (filename, drive, S9xGetSnapshotDirectory (), 
+			       def, ext);
+		    if (S9xSPCDump (filename))
+			sprintf (InfoString, "%s.%s saved", def, ext);
+		    else
+			sprintf (InfoString, "%s.%s not saved (%s)", def, ext,
+				 strerror (errno));
+		    
+		    S9xSetInfoString (InfoString);
 		    break;
 		}
-#endif
 		// Fall...
 	    case XK_F2:
 		if (event.type == KeyPress && 
 		    (event.xkey.state & (Mod1Mask | ControlMask)))
 		{
-		    S9xLoadSnapshot (S9xChooseFilename (TRUE));
+		    S9xUnfreezeGame (S9xChooseFilename (TRUE));
 		    break;
 		}
 		// Fall...
@@ -965,7 +1136,7 @@ void S9xProcessEvents (bool8 block)
 		if (event.type == KeyPress && 
 		    (event.xkey.state & (Mod1Mask | ControlMask)))
 		{
-		    Snapshot (S9xChooseFilename (FALSE));
+		    S9xFreezeGame (S9xChooseFilename (FALSE));
 		    break;
 		}
 		// Fall...
@@ -1002,13 +1173,25 @@ void S9xProcessEvents (bool8 block)
 			sprintf (filename, "%s%s%s.%03d",
 				 S9xGetSnapshotDirectory (), SLASH_STR, def,
 				 key - XK_F1);
-			S9xLoadSnapshot (filename);
+			if (S9xLoadSnapshot (filename))
+			{
+			    sprintf (InfoString, "%s.%03d loaded", def, key - XK_F1);
+			    S9xSetInfoString (InfoString);
+			}
 		    }
 		    else
 		    if (event.xkey.state & (Mod1Mask | ControlMask))
 		    {
 			if (key >= XK_F4)
 			    S9xToggleSoundChannel (key - XK_F4);
+			if (key == XK_F12)
+			    S9xSetInfoString ("All sound channels on");
+			else
+			{
+			    sprintf (InfoString, "Sound channel %d %s", key - XK_F4,
+				     (so.sound_switch & (1 << (key - XK_F4))) ? "on" : "off");
+			    S9xSetInfoString (InfoString);
+			}
 		    }
 		    else
 		    {
@@ -1022,6 +1205,8 @@ void S9xProcessEvents (bool8 block)
 			sprintf (filename, "%s%s%s.%03d",
 				 S9xGetSnapshotDirectory (), SLASH_STR, def,
 				 key - XK_F1);
+			sprintf (InfoString, "%s.%03d saved", def, key - XK_F1);
+			S9xSetInfoString (InfoString);
 			Snapshot (filename);
 		    }
 		}
@@ -1065,6 +1250,13 @@ void S9xProcessEvents (bool8 block)
 		IPPU.FrameSkip = Settings.SkipFrames;
 		SetupImage ();
 	    }
+#ifdef USE_DGA_EXTENSION
+	    if (xf86data.start_full_screen)
+	    {
+		xf86data.start_full_screen = FALSE;
+		S9xSwitchToFullScreen (TRUE);
+	    }
+#endif
 	    break;
 #if 0
 	case ButtonPress:
@@ -1080,6 +1272,12 @@ void S9xProcessEvents (bool8 block)
 
 void S9xPutImage (int snes_width, int snes_height)
 {
+#ifdef USE_GLIDE
+    if (Settings.GlideEnable)
+	S9xGlidePutImage (snes_width, snes_height);
+    else
+    {
+#endif
     bool8 scaled = FALSE;
     bool8 done = FALSE;
     int width, height;
@@ -1091,17 +1289,12 @@ void S9xPutImage (int snes_width, int snes_height)
     {
 	if (snes_width == 512 && snes_height > 240)
 	{
-	    if (ImageNeedsScaling)
+	    OutputScreen = GFX.Screen;
+	    OutputPitch = GFX.Pitch;
+
+	    if (!ImageNeedsScaling)
 	    {
-		OutputScreen = InterpolateScreen;
-		OutputPitch = 514 * 2;
-	    }
-	    else
-	    {
-		OutputScreen = GFX.Screen;
-		OutputPitch = GFX.Pitch;
-		    OutputScreen = (uint8 *) image->data;
-		    OutputPitch = image->bytes_per_line;
+		memmove (image_data, GFX.Screen, GFX.Pitch * snes_height);
 	    }
 	}
 	else
@@ -1226,6 +1419,7 @@ void S9xPutImage (int snes_width, int snes_height)
 #ifdef USE_DGA_EXTENSION
     }
 #endif
+
     last_snes_width = snes_width;
     last_snes_height = snes_height;
 
@@ -1296,10 +1490,18 @@ void S9xPutImage (int snes_width, int snes_height)
 	mouse_buttons = ((mask & 0x100) >> 8) | ((mask & 0x200) >> 7) |
 			((mask & 0x400) >> 9) | ((mask & 0x800) >> 8);
     }
+#ifdef USE_GLIDE
+    }
+#endif
 }
 
 void S9xSetPalette ()
 {
+#ifdef USE_GLIDE
+    if (Settings.GlideEnable)
+	return;
+#endif
+
     int i;
 
     if (grayscale)
@@ -1462,13 +1664,13 @@ void Convert8To24 (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint32 *d = (uint32 *) (image_data +
 					     y * bytes_per_line);
 	    register uint8 *s = OutputScreen + y * OutputPitch;
 
-	    for (register x = 0; x < width; x++)
+	    for (register int x = 0; x < width; x++)
 	    {
 		uint32 pixel = PPU.CGDATA [*s++];
 		*d++ = (((pixel & 0x1f) * brightness) << RedShift) |
@@ -1524,13 +1726,13 @@ void Convert16To24 (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint32 *d = (uint32 *) (image_data +
 					     y * bytes_per_line);
 	    register uint16 *s = (uint16 *) (OutputScreen + y * OutputPitch);
 
-	    for (register x = 0; x < width; x++)
+	    for (register int x = 0; x < width; x++)
 	    {
 		uint32 pixel = *s++;
 		*d++ = (((pixel >> 11) & 0x1f) << (RedShift + 3)) |
@@ -1592,7 +1794,7 @@ void Convert8To24Packed (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint8 *d = (uint8 *) (image_data + y * bytes_per_line);
 	    register uint8 *s = OutputScreen + y * OutputPitch;
@@ -1604,7 +1806,7 @@ void Convert8To24Packed (int width, int height)
 #endif
 	    {
 		// Order is RGB
-		for (register x = 0; x < width; x++)
+		for (register int x = 0; x < width; x++)
 		{
 		    uint16 pixel = PPU.CGDATA [*s++];
 		    *d++ = levels [(pixel & 0x1f)];
@@ -1615,7 +1817,7 @@ void Convert8To24Packed (int width, int height)
 	    else
 	    {
 		// Order is BGR
-		for (register x = 0; x < width; x++)
+		for (register int x = 0; x < width; x++)
 		{
 		    uint16 pixel = PPU.CGDATA [*s++];
 		    *d++ = levels [((pixel >> 10) & 0x1f)];
@@ -1698,7 +1900,7 @@ void Convert16To24Packed (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint8 *d = (uint8 *) (image_data +
 					 y * bytes_per_line);
@@ -1711,7 +1913,7 @@ void Convert16To24Packed (int width, int height)
 #endif
 	    {
 		// Order is RGB
-		for (register x = 0; x < width; x++)
+		for (register int x = 0; x < width; x++)
 		{
 		    uint32 pixel = *s++;
 		    *d++ = (pixel >> (11 - 3)) & 0xf8;
@@ -1722,7 +1924,7 @@ void Convert16To24Packed (int width, int height)
 	    else
 	    {
 		// Order is BGR
-		for (register x = 0; x < width; x++)
+		for (register int x = 0; x < width; x++)
 		{
 		    uint32 pixel = *s++;
 		    *d++ = (pixel & 0x1f) << 3;
@@ -1805,12 +2007,12 @@ void Convert16To8 (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint8 *d = (uint8 *) image_data + y * bytes_per_line;
-	    register uint16 *s = (uint16 *) (OutputScreen + y * GFX.RealPitch);
+	    register uint16 *s = (uint16 *) (OutputScreen + y * OutputPitch);
 
-	    for (register x = 0; x < width; x++)
+	    for (register int x = 0; x < width; x++)
 		*d++ = palette [*s++];
 	}
     }
@@ -1862,12 +2064,12 @@ void Convert8To16 (int width, int height)
     if (!ImageNeedsScaling)
     {
 	// Convert
-	for (register y = 0; y < height; y++)
+	for (register int y = 0; y < height; y++)
 	{
 	    register uint16 *d = (uint16 *) (image_data + y * bytes_per_line);
 	    register uint8 *s = OutputScreen + y * OutputPitch;
 
-	    for (register x = 0; x < width; x++)
+	    for (register int x = 0; x < width; x++)
 	    {
 		uint32 pixel = PPU.CGDATA [*s++];
 		*d++ = (levels [pixel & 0x1f] << RedShift) |
@@ -1996,7 +2198,11 @@ void S9xParseDisplayArg (char **argv, int &ind, int)
     if (strcasecmp (argv [ind], "-nms") == 0 ||
 	strcasecmp (argv [ind], "-nomodeswitch") == 0)
 	xf86data.no_mode_switch = TRUE;
-#endif	
+#endif
+    else
+    if (strcasecmp (argv [ind], "-fs") == 0 ||
+	strcasecmp (argv [ind], "-fullscreen") == 0)
+	xf86data.start_full_screen = TRUE;
 #endif
     else
 	S9xUsage ();
@@ -2013,7 +2219,14 @@ int S9xMinCommandLineArgs ()
 
 void S9xMessage (int /*type*/, int /*number*/, const char *message)
 {
+#define MAX_MESSAGE_LEN (36 * 3)
+
+    static char buffer [MAX_MESSAGE_LEN + 1];
+
     fprintf (stdout, "%s\n", message);
+    strncpy (buffer, message, MAX_MESSAGE_LEN);
+    buffer [MAX_MESSAGE_LEN] = 0;
+    S9xSetInfoString (buffer);
 }
 
 void TVMode (int width, int height)
@@ -2155,7 +2368,7 @@ void TVMode (int width, int height)
 extern "C" {
 #include <sys/ioctl.h>
 
-// Select seems to be broken in 2.0.x kernels - if a signal interrupts a
+// Select seems to be broken in 2.x.x kernels - if a signal interrupts a
 // select system call with a zero timeout, the select call is restarted but
 // with an infinite timeout! The call will block until data arrives on the
 // selected fd(s).
@@ -2327,3 +2540,23 @@ void S9xSwitchToFullScreen (bool8 enable)
     }
 }
 #endif
+
+void S9xDisplayStateChange (const char *str, bool8 on)
+{
+    static char string [100];
+
+    sprintf (string, "%s %s", str, on ? "on" : "off");
+    S9xSetInfoString (string);
+}
+
+#ifdef USE_GLIDE
+void S9xSwitchToGlideMode (bool8 enable)
+{
+    S9xGlideEnable (enable);
+    if (Settings.GlideEnable)
+	XGrabKeyboard (display, window, True, GrabModeAsync, GrabModeAsync,
+		       CurrentTime);
+    else
+	XUngrabKeyboard (display, CurrentTime);
+}
+#endif
